#!/usr/bin/python
#
# Copyright (c) 2023, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: docker_compose_v2_run

short_description: Run command in a new container of a Compose service

version_added: 3.13.0

description:
  - Uses Docker Compose to run a command in a new container for a service.
  - This encapsulates C(docker compose run).
extends_documentation_fragment:
  - community.docker._compose_v2
  - community.docker._compose_v2.minimum_version
  - community.docker._docker.cli_documentation
  - community.docker._attributes
  - community.docker._attributes.actiongroup_docker

attributes:
  check_mode:
    support: none
  diff_mode:
    support: none
  idempotent:
    support: N/A
    details:
      - Whether the executed command is idempotent depends on the command.

options:
  service:
    description:
      - The service to run the command in.
    type: str
    required: true
  argv:
    type: list
    elements: str
    description:
      - The command to execute.
      - Since this is a list of arguments, no quoting is needed.
      - O(argv) or O(command) are mutually exclusive.
  command:
    type: str
    description:
      - The command to execute.
      - O(argv) or O(command) are mutually exclusive.
  build:
    description:
      - Build image before starting container.
      - Note that building can insert information into RV(stdout) or RV(stderr).
    type: bool
    default: false
  cap_add:
    description:
      - Linux capabilities to add to the container.
    type: list
    elements: str
  cap_drop:
    description:
      - Linux capabilities to drop from the container.
    type: list
    elements: str
  entrypoint:
    description:
      - Override the entrypoint of the container image.
    type: str
  interactive:
    description:
      - Whether to keep STDIN open even if not attached.
    type: bool
    default: true
  labels:
    description:
      - Add or override labels to the container.
    type: list
    elements: str
  name:
    description:
      - Assign a name to the container.
    type: str
  no_deps:
    description:
      - Do not start linked services.
    type: bool
    default: false
  publish:
    description:
      - Publish a container's port(s) to the host.
    type: list
    elements: str
  quiet_pull:
    description:
      - Pull without printing progress information.
      - Note that pulling can insert information into RV(stdout) or RV(stderr).
    type: bool
    default: false
  remove_orphans:
    description:
      - Remove containers for services not defined in the Compose file.
    type: bool
    default: false
  cleanup:
    description:
      - Automatically remove th econtainer when it exits.
      - Corresponds to the C(--rm) option of C(docker compose run).
    type: bool
    default: false
  service_ports:
    description:
      - Run command with all service's ports enabled and mapped to the host.
    type: bool
    default: false
  use_aliases:
    description:
      - Use the service's network C(useAliases) in the network(s) the container connects to.
    type: bool
    default: false
  volumes:
    description:
      - Bind mount one or more volumes.
    type: list
    elements: str
  chdir:
    type: str
    description:
      - The directory to run the command in.
  detach:
    description:
      - Whether to run the command synchronously (O(detach=false), default) or asynchronously (O(detach=true)).
      - If set to V(true), O(stdin) cannot be provided, and the return values RV(stdout), RV(stderr), and RV(rc) are not returned.
        Instead, the return value RV(container_id) is provided.
    type: bool
    default: false
  user:
    type: str
    description:
      - If specified, the user to execute this command with.
  stdin:
    type: str
    description:
      - Set the stdin of the command directly to the specified value.
      - Can only be used if O(detach=false).
  stdin_add_newline:
    type: bool
    default: true
    description:
      - If set to V(true), appends a newline to O(stdin).
  strip_empty_ends:
    type: bool
    default: true
    description:
      - Strip empty lines from the end of stdout/stderr in result.
  tty:
    type: bool
    default: true
    description:
      - Whether to allocate a TTY.
  env:
    description:
      - Dictionary of environment variables with their respective values to be passed to the command ran inside the container.
      - Values which might be parsed as numbers, booleans or other types by the YAML parser must be quoted (for example V("true"))
        in order to avoid data loss.
      - Please note that if you are passing values in with Jinja2 templates, like V("{{ value }}"), you need to add V(| string)
        to prevent Ansible to convert strings such as V("true") back to booleans. The correct way is to use V("{{ value |
        string }}").
    type: dict

author:
  - Felix Fontein (@felixfontein)

seealso:
  - module: community.docker.docker_compose_v2

notes:
  - If you need to evaluate environment variables of the container in O(command) or O(argv), you need to pass the command
    through a shell, like O(command=/bin/sh -c "echo $ENV_VARIABLE"). The same needs to be done in case you want to use glob patterns
    or other shell features such as redirects.
"""

EXAMPLES = r"""
---
- name: Run a simple command (command)
  community.docker.docker_compose_v2_run:
    service: foo
    command: /bin/bash -c "ls -lah"
    chdir: /root
  register: result

- name: Print stdout
  ansible.builtin.debug:
    var: result.stdout

- name: Run a simple command (argv)
  community.docker.docker_compose_v2_run:
    service: foo
    argv:
      - /bin/bash
      - "-c"
      - "ls -lah > /dev/stderr"
    chdir: /root
  register: result

- name: Print stderr lines
  ansible.builtin.debug:
    var: result.stderr_lines
"""

RETURN = r"""
container_id:
  type: str
  returned: success and O(detach=true)
  description:
    - The ID of the created container.
stdout:
  type: str
  returned: success and O(detach=false)
  description:
    - The standard output of the container command.
stderr:
  type: str
  returned: success and O(detach=false)
  description:
    - The standard error output of the container command.
rc:
  type: int
  returned: success and O(detach=false)
  sample: 0
  description:
    - The exit code of the command.
"""

import shlex
import traceback
import typing as t

from ansible.module_utils.common.text.converters import to_text

from ansible_collections.community.docker.plugins.module_utils._common_cli import (
    AnsibleModuleDockerClient,
    DockerException,
)
from ansible_collections.community.docker.plugins.module_utils._compose_v2 import (
    BaseComposeManager,
    common_compose_argspec_ex,
)


class ExecManager(BaseComposeManager):
    def __init__(self, client: AnsibleModuleDockerClient) -> None:
        super().__init__(client)
        parameters = self.client.module.params

        self.service: str = parameters["service"]
        self.build: bool = parameters["build"]
        self.cap_add: list[str] | None = parameters["cap_add"]
        self.cap_drop: list[str] | None = parameters["cap_drop"]
        self.entrypoint: str | None = parameters["entrypoint"]
        self.interactive: bool = parameters["interactive"]
        self.labels: list[str] | None = parameters["labels"]
        self.name: str | None = parameters["name"]
        self.no_deps: bool = parameters["no_deps"]
        self.publish: list[str] | None = parameters["publish"]
        self.quiet_pull: bool = parameters["quiet_pull"]
        self.remove_orphans: bool = parameters["remove_orphans"]
        self.do_cleanup: bool = parameters["cleanup"]
        self.service_ports: bool = parameters["service_ports"]
        self.use_aliases: bool = parameters["use_aliases"]
        self.volumes: list[str] | None = parameters["volumes"]
        self.chdir: str | None = parameters["chdir"]
        self.detach: bool = parameters["detach"]
        self.user: str | None = parameters["user"]
        self.stdin: str | None = parameters["stdin"]
        self.strip_empty_ends: bool = parameters["strip_empty_ends"]
        self.tty: bool = parameters["tty"]
        self.env: dict[str, t.Any] | None = parameters["env"]

        self.argv: list[str]
        if parameters["command"] is not None:
            self.argv = shlex.split(parameters["command"])
        else:
            self.argv = parameters["argv"]

        if self.detach and self.stdin is not None:
            self.fail("If detach=true, stdin cannot be provided.")

        stdin_add_newline: bool = parameters["stdin_add_newline"]
        if self.stdin is not None and stdin_add_newline:
            self.stdin += "\n"

        if self.env is not None:
            for name, value in self.env.items():
                if not isinstance(value, str):
                    self.fail(
                        "Non-string value found for env option. Ambiguous env options must be "
                        "wrapped in quotes to avoid them being interpreted when directly specified "
                        "in YAML, or explicitly converted to strings when the option is templated. "
                        f"Key: {name}"
                    )

    def get_run_cmd(self, dry_run: bool) -> list[str]:
        args = self.get_base_args(plain_progress=True) + ["run"]
        if self.build:
            args.append("--build")
        if self.cap_add:
            for cap in self.cap_add:
                args.extend(["--cap-add", cap])
        if self.cap_drop:
            for cap in self.cap_drop:
                args.extend(["--cap-drop", cap])
        if self.entrypoint is not None:
            args.extend(["--entrypoint", self.entrypoint])
        if not self.interactive:
            args.append("--no-interactive")
        if self.labels:
            for label in self.labels:
                args.extend(["--label", label])
        if self.name is not None:
            args.extend(["--name", self.name])
        if self.no_deps:
            args.append("--no-deps")
        if self.publish:
            for publish in self.publish:
                args.extend(["--publish", publish])
        if self.quiet_pull:
            args.append("--quiet-pull")
        if self.remove_orphans:
            args.append("--remove-orphans")
        if self.do_cleanup:
            args.append("--rm")
        if self.service_ports:
            args.append("--service-ports")
        if self.use_aliases:
            args.append("--use-aliases")
        if self.volumes:
            for volume in self.volumes:
                args.extend(["--volume", volume])
        if self.chdir is not None:
            args.extend(["--workdir", self.chdir])
        if self.detach:
            args.extend(["--detach"])
        if self.user is not None:
            args.extend(["--user", self.user])
        if not self.tty:
            args.append("--no-TTY")
        if self.env:
            for name, value in list(self.env.items()):
                args.append("--env")
                args.append(f"{name}={value}")
        args.append("--")
        args.append(self.service)
        if self.argv:
            args.extend(self.argv)
        return args

    def run(self) -> dict[str, t.Any]:
        args = self.get_run_cmd(self.check_mode)
        kwargs: dict[str, t.Any] = {
            "cwd": self.project_src,
        }
        if self.stdin is not None:
            kwargs["data"] = self.stdin.encode("utf-8")
        if self.detach:
            kwargs["check_rc"] = True
        rc, stdout_b, stderr_b = self.client.call_cli(*args, **kwargs)
        if self.detach:
            return {
                "container_id": to_text(stdout_b.strip()),
            }
        stdout = to_text(stdout_b)
        stderr = to_text(stderr_b)
        if self.strip_empty_ends:
            stdout = stdout.rstrip("\r\n")
            stderr = stderr.rstrip("\r\n")
        return {
            "changed": True,
            "rc": rc,
            "stdout": stdout,
            "stderr": stderr,
        }


def main() -> None:
    argument_spec = {
        "service": {"type": "str", "required": True},
        "argv": {"type": "list", "elements": "str"},
        "command": {"type": "str"},
        "build": {"type": "bool", "default": False},
        "cap_add": {"type": "list", "elements": "str"},
        "cap_drop": {"type": "list", "elements": "str"},
        "entrypoint": {"type": "str"},
        "interactive": {"type": "bool", "default": True},
        "labels": {"type": "list", "elements": "str"},
        "name": {"type": "str"},
        "no_deps": {"type": "bool", "default": False},
        "publish": {"type": "list", "elements": "str"},
        "quiet_pull": {"type": "bool", "default": False},
        "remove_orphans": {"type": "bool", "default": False},
        "cleanup": {"type": "bool", "default": False},
        "service_ports": {"type": "bool", "default": False},
        "use_aliases": {"type": "bool", "default": False},
        "volumes": {"type": "list", "elements": "str"},
        "chdir": {"type": "str"},
        "detach": {"type": "bool", "default": False},
        "user": {"type": "str"},
        "stdin": {"type": "str"},
        "stdin_add_newline": {"type": "bool", "default": True},
        "strip_empty_ends": {"type": "bool", "default": True},
        "tty": {"type": "bool", "default": True},
        "env": {"type": "dict"},
    }
    argspec_ex = common_compose_argspec_ex()
    argument_spec.update(argspec_ex.pop("argspec"))

    client = AnsibleModuleDockerClient(
        argument_spec=argument_spec,
        supports_check_mode=False,
        needs_api_version=False,
        **argspec_ex,
    )

    try:
        manager = ExecManager(client)
        result = manager.run()
        manager.cleanup()
        client.module.exit_json(**result)
    except DockerException as e:
        client.fail(
            f"An unexpected Docker error occurred: {e}",
            exception=traceback.format_exc(),
        )


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: ecs_task
version_added: 1.0.0
short_description: Run, start or stop a task in ECS
description:
    - Creates or deletes instances of task definitions.
author:
    - Mark Chance (@Java1Guy)
options:
    operation:
        description:
            - Which task operation to execute.
            - When I(operation=run) I(task_definition) must be set.
            - When I(operation=start) both I(task_definition) and I(container_instances) must be set.
            - When I(operation=stop) both I(task_definition) and I(task) must be set.
        required: True
        choices: ['run', 'start', 'stop']
        type: str
    cluster:
        description:
            - The name of the cluster to run the task on.
            - If not specified, the cluster name will be C(default).
        required: False
        type: str
        default: 'default'
    task_definition:
        description:
            - The task definition to start, run or stop.
        required: False
        type: str
    overrides:
        description:
            - A dictionary of values to pass to the new instances.
        required: False
        type: dict
    count:
        description:
            - How many new instances to start.
        required: False
        type: int
    task:
        description:
            - The ARN of the task to stop.
        required: False
        type: str
    container_instances:
        description:
            - The list of container instances on which to deploy the task.
        required: False
        type: list
        elements: str
    started_by:
        description:
            - A value showing who or what started the task (for informational purposes).
        required: False
        type: str
    network_configuration:
        description:
          - Network configuration of the service. Only applicable for task definitions created with I(network_mode=awsvpc).
        type: dict
        suboptions:
            assign_public_ip:
                description: Whether the task's elastic network interface receives a public IP address.
                type: bool
                version_added: 1.5.0
            subnets:
                description: A list of subnet IDs to which the task is attached.
                type: list
                elements: str
            security_groups:
                description: A list of group names or group IDs for the task.
                type: list
                elements: str
    launch_type:
        description:
          - The launch type on which to run your service.
        required: false
        choices: ["EC2", "FARGATE"]
        type: str
    tags:
        type: dict
        description:
          - Tags that will be added to ecs tasks on start and run
        required: false
        aliases: ['resource_tags']
    wait:
        description:
          - Whether or not to wait for the desired state.
        type: bool
        default: false
        version_added: 4.1.0
extends_documentation_fragment:
    - amazon.aws.common.modules
    - amazon.aws.region.modules
    - amazon.aws.boto3
"""

EXAMPLES = r"""
# Simple example of run task
- name: Run task
  community.aws.ecs_task:
    operation: run
    cluster: console-sample-app-static-cluster
    task_definition: console-sample-app-static-taskdef
    count: 1
    started_by: ansible_user
  register: task_output

# Simple example of start task

- name: Start a task
  community.aws.ecs_task:
    operation: start
    cluster: console-sample-app-static-cluster
    task_definition: console-sample-app-static-taskdef
    task: "arn:aws:ecs:us-west-2:123456789012:task/3f8353d1-29a8-4689-bbf6-ad79937ffe8a"
    tags:
      resourceName: a_task_for_ansible_to_run
      type: long_running_task
      network: internal
      version: 1.4
    container_instances:
      - arn:aws:ecs:us-west-2:123456789012:container-instance/79c23f22-876c-438a-bddf-55c98a3538a8
    started_by: ansible_user
    network_configuration:
      subnets:
        - subnet-abcd1234
      security_groups:
        - sg-aaaa1111
        - my_security_group
  register: task_output

- name: RUN a task on Fargate
  community.aws.ecs_task:
    operation: run
    cluster: console-sample-app-static-cluster
    task_definition: console-sample-app-static-taskdef
    task: "arn:aws:ecs:us-west-2:123456789012:task/3f8353d1-29a8-4689-bbf6-ad79937ffe8a"
    started_by: ansible_user
    launch_type: FARGATE
    network_configuration:
      subnets:
        - subnet-abcd1234
      security_groups:
        - sg-aaaa1111
        - my_security_group
  register: task_output

- name: RUN a task on Fargate with public ip assigned
  community.aws.ecs_task:
    operation: run
    count: 2
    cluster: console-sample-app-static-cluster
    task_definition: console-sample-app-static-taskdef
    task: "arn:aws:ecs:us-west-2:123456789012:task/3f8353d1-29a8-4689-bbf6-ad79937ffe8a"
    started_by: ansible_user
    launch_type: FARGATE
    network_configuration:
      assign_public_ip: true
      subnets:
        - subnet-abcd1234
  register: task_output

- name: Stop a task
  community.aws.ecs_task:
    operation: stop
    cluster: console-sample-app-static-cluster
    task_definition: console-sample-app-static-taskdef
    task: "arn:aws:ecs:us-west-2:123456789012:task/3f8353d1-29a8-4689-bbf6-ad79937ffe8a"
"""

RETURN = r"""
task:
    description: details about the task that was started
    returned: success
    type: complex
    contains:
        taskArn:
            description: The Amazon Resource Name (ARN) that identifies the task.
            returned: always
            type: str
        clusterArn:
            description: The Amazon Resource Name (ARN) of the of the cluster that hosts the task.
            returned: only when details is true
            type: str
        taskDefinitionArn:
            description: The Amazon Resource Name (ARN) of the task definition.
            returned: only when details is true
            type: str
        containerInstanceArn:
            description: The Amazon Resource Name (ARN) of the container running the task.
            returned: only when details is true
            type: str
        overrides:
            description: The container overrides set for this task.
            returned: only when details is true
            type: list
            elements: dict
        lastStatus:
            description: The last recorded status of the task.
            returned: only when details is true
            type: str
        desiredStatus:
            description: The desired status of the task.
            returned: only when details is true
            type: str
        containers:
            description: The container details.
            returned: only when details is true
            type: list
            elements: dict
        startedBy:
            description: The used who started the task.
            returned: only when details is true
            type: str
        stoppedReason:
            description: The reason why the task was stopped.
            returned: only when details is true
            type: str
        createdAt:
            description: The timestamp of when the task was created.
            returned: only when details is true
            type: str
        startedAt:
            description: The timestamp of when the task was started.
            returned: only when details is true
            type: str
        stoppedAt:
            description: The timestamp of when the task was stopped.
            returned: only when details is true
            type: str
        launchType:
            description: The launch type on which to run your task.
            returned: always
            type: str
"""

try:
    import botocore
except ImportError:
    pass  # caught by AnsibleAWSModule

from ansible_collections.amazon.aws.plugins.module_utils.ec2 import get_ec2_security_group_ids_from_names
from ansible_collections.amazon.aws.plugins.module_utils.tagging import ansible_dict_to_boto3_tag_list

from ansible_collections.community.aws.plugins.module_utils.modules import AnsibleCommunityAWSModule as AnsibleAWSModule


class EcsExecManager:
    """Handles ECS Tasks"""

    def __init__(self, module):
        self.module = module
        self.ecs = module.client("ecs")
        self.ec2 = module.client("ec2")

    def format_network_configuration(self, network_config):
        result = dict()
        if "subnets" in network_config:
            result["subnets"] = network_config["subnets"]
        else:
            self.module.fail_json(msg="Network configuration must include subnets")
        if "security_groups" in network_config:
            groups = network_config["security_groups"]
            if any(not sg.startswith("sg-") for sg in groups):
                try:
                    vpc_id = self.ec2.describe_subnets(SubnetIds=[result["subnets"][0]])["Subnets"][0]["VpcId"]
                    groups = get_ec2_security_group_ids_from_names(groups, self.ec2, vpc_id)
                except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
                    self.module.fail_json_aws(e, msg="Couldn't look up security groups")
            result["securityGroups"] = groups
        if "assign_public_ip" in network_config:
            if network_config["assign_public_ip"] is True:
                result["assignPublicIp"] = "ENABLED"
            else:
                result["assignPublicIp"] = "DISABLED"

        return dict(awsvpcConfiguration=result)

    def list_tasks(self, cluster_name, service_name, status):
        response = self.ecs.list_tasks(
            cluster=cluster_name,
            family=service_name,
            desiredStatus=status,
        )
        if len(response["taskArns"]) > 0:
            for c in response["taskArns"]:
                if c.endswith(service_name):
                    return c
        return None

    def run_task(self, cluster, task_definition, overrides, count, startedBy, launch_type, tags):
        if overrides is None:
            overrides = dict()
        params = dict(
            cluster=cluster, taskDefinition=task_definition, overrides=overrides, count=count, startedBy=startedBy
        )
        if self.module.params["network_configuration"]:
            params["networkConfiguration"] = self.format_network_configuration(
                self.module.params["network_configuration"]
            )
        if launch_type:
            params["launchType"] = launch_type
        if tags:
            params["tags"] = ansible_dict_to_boto3_tag_list(tags, "key", "value")

            # TODO: need to check if long arn format enabled.
        try:
            response = self.ecs.run_task(**params)
        except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
            self.module.fail_json_aws(e, msg="Couldn't run task")
        # include tasks and failures
        return response["tasks"]

    def start_task(self, cluster, task_definition, overrides, container_instances, startedBy, tags):
        args = dict()
        if cluster:
            args["cluster"] = cluster
        if task_definition:
            args["taskDefinition"] = task_definition
        if overrides:
            args["overrides"] = overrides
        if container_instances:
            args["containerInstances"] = container_instances
        if startedBy:
            args["startedBy"] = startedBy
        if self.module.params["network_configuration"]:
            args["networkConfiguration"] = self.format_network_configuration(
                self.module.params["network_configuration"]
            )
        if tags:
            args["tags"] = ansible_dict_to_boto3_tag_list(tags, "key", "value")
        try:
            response = self.ecs.start_task(**args)
        except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
            self.module.fail_json_aws(e, msg="Couldn't start task")
        # include tasks and failures
        return response["tasks"]

    def stop_task(self, cluster, task):
        response = self.ecs.stop_task(cluster=cluster, task=task)
        return response["task"]

    def ecs_task_long_format_enabled(self):
        account_support = self.ecs.list_account_settings(name="taskLongArnFormat", effectiveSettings=True)
        return account_support["settings"][0]["value"] == "enabled"


def main():
    argument_spec = dict(
        operation=dict(required=True, choices=["run", "start", "stop"]),
        cluster=dict(required=False, type="str", default="default"),  # R S P
        task_definition=dict(required=False, type="str"),  # R* S*
        overrides=dict(required=False, type="dict"),  # R S
        count=dict(required=False, type="int"),  # R
        task=dict(required=False, type="str"),  # P*
        container_instances=dict(required=False, type="list", elements="str"),  # S*
        started_by=dict(required=False, type="str"),  # R S
        network_configuration=dict(required=False, type="dict"),
        launch_type=dict(required=False, choices=["EC2", "FARGATE"]),
        tags=dict(required=False, type="dict", aliases=["resource_tags"]),
        wait=dict(required=False, default=False, type="bool"),
    )

    module = AnsibleAWSModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ("launch_type", "FARGATE", ["network_configuration"]),
            ("operation", "run", ["task_definition"]),
            ("operation", "start", ["task_definition", "container_instances"]),
            ("operation", "stop", ["task_definition", "task"]),
        ],
    )

    # Validate Inputs
    if module.params["operation"] == "run":
        task_to_list = module.params["task_definition"]
        status_type = "RUNNING"

    if module.params["operation"] == "start":
        task_to_list = module.params["task"]
        status_type = "RUNNING"

    if module.params["operation"] == "stop":
        task_to_list = module.params["task_definition"]
        status_type = "STOPPED"

    service_mgr = EcsExecManager(module)

    if module.params["tags"]:
        if not service_mgr.ecs_task_long_format_enabled():
            module.fail_json(msg="Cannot set task tags: long format task arns are required to set tags")

    existing = service_mgr.list_tasks(module.params["cluster"], task_to_list, status_type)

    results = dict(changed=False)
    if module.params["operation"] == "run":
        if existing:
            # TBD - validate the rest of the details
            results["task"] = existing
        else:
            if not module.check_mode:
                # run_task returns a list of tasks created
                tasks = service_mgr.run_task(
                    module.params["cluster"],
                    module.params["task_definition"],
                    module.params["overrides"],
                    module.params["count"],
                    module.params["started_by"],
                    module.params["launch_type"],
                    module.params["tags"],
                )

                # Wait for task(s) to be running prior to exiting
                if module.params["wait"]:
                    waiter = service_mgr.ecs.get_waiter("tasks_running")
                    try:
                        waiter.wait(
                            tasks=[task["taskArn"] for task in tasks],
                            cluster=module.params["cluster"],
                        )
                    except botocore.exceptions.WaiterError as e:
                        module.fail_json_aws(e, "Timeout waiting for tasks to run")

                results["task"] = tasks

            results["changed"] = True

    elif module.params["operation"] == "start":
        if existing:
            # TBD - validate the rest of the details
            results["task"] = existing
        else:
            if not module.check_mode:
                results["task"] = service_mgr.start_task(
                    module.params["cluster"],
                    module.params["task_definition"],
                    module.params["overrides"],
                    module.params["container_instances"],
                    module.params["started_by"],
                    module.params["tags"],
                )

            results["changed"] = True

    elif module.params["operation"] == "stop":
        if existing:
            results["task"] = existing
        else:
            if not module.check_mode:
                # it exists, so we should delete it and mark changed.
                # return info about the cluster deleted
                results["task"] = service_mgr.stop_task(module.params["cluster"], module.params["task"])

                # Wait for task to be stopped prior to exiting
                if module.params["wait"]:
                    waiter = service_mgr.ecs.get_waiter("tasks_stopped")
                    try:
                        waiter.wait(
                            tasks=[module.params["task"]],
                            cluster=module.params["cluster"],
                        )
                    except botocore.exceptions.WaiterError as e:
                        module.fail_json_aws(e, "Timeout waiting for task to stop")

            results["changed"] = True

    module.exit_json(**results)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: networks_switch_link_aggregations
short_description: Resource module for networks _switch _link _aggregations
description:
  - Manage operations create, update and delete of the resource networks _switch _link
    _aggregations.
  - Create a link aggregation group.
  - Split a link aggregation group into separate ports.
  - Update a link aggregation group.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module
author: Francisco Munoz (@fmunoz)
options:
  linkAggregationId:
    description: LinkAggregationId path parameter. Link aggregation ID.
    type: str
  networkId:
    description: NetworkId path parameter. Network ID.
    type: str
  switchPorts:
    description: Array of switch or stack ports for creating aggregation group. Minimum
      2 and maximum 8 ports are supported.
    elements: dict
    suboptions:
      portId:
        description: Port identifier of switch port. For modules, the identifier is
          "SlotNumber_ModuleType_PortNumber" (Ex "1_8X10G_1"), otherwise it is just
          the port number (Ex "8").
        type: str
      serial:
        description: Serial number of the switch.
        type: str
    type: list
  switchProfilePorts:
    description: Array of switch profile ports for creating aggregation group. Minimum
      2 and maximum 8 ports are supported.
    elements: dict
    suboptions:
      portId:
        description: Port identifier of switch port. For modules, the identifier is
          "SlotNumber_ModuleType_PortNumber" (Ex "1_8X10G_1"), otherwise it is just
          the port number (Ex "8").
        type: str
      profile:
        description: Profile identifier.
        type: str
    type: list
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for switch createNetworkSwitchLinkAggregation
    description: Complete reference of the createNetworkSwitchLinkAggregation API.
    link: https://developer.cisco.com/meraki/api-v1/#!create-network-switch-link-aggregation
  - name: Cisco Meraki documentation for switch deleteNetworkSwitchLinkAggregation
    description: Complete reference of the deleteNetworkSwitchLinkAggregation API.
    link: https://developer.cisco.com/meraki/api-v1/#!delete-network-switch-link-aggregation
  - name: Cisco Meraki documentation for switch updateNetworkSwitchLinkAggregation
    description: Complete reference of the updateNetworkSwitchLinkAggregation API.
    link: https://developer.cisco.com/meraki/api-v1/#!update-network-switch-link-aggregation
notes:
  - SDK Method used are
    switch.Switch.create_network_switch_link_aggregation,
    switch.Switch.delete_network_switch_link_aggregation,
    switch.Switch.update_network_switch_link_aggregation,
  - Paths used are
    post /networks/{networkId}/switch/linkAggregations,
    delete /networks/{networkId}/switch/linkAggregations/{linkAggregationId},
    put /networks/{networkId}/switch/linkAggregations/{linkAggregationId},
"""

EXAMPLES = r"""
- name: Create
  cisco.meraki.networks_switch_link_aggregations:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    networkId: string
    switchPorts:
      - portId: '1'
        serial: Q234-ABCD-0001
    switchProfilePorts:
      - portId: '2'
        profile: '1234'
- name: Update by id
  cisco.meraki.networks_switch_link_aggregations:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    linkAggregationId: string
    networkId: string
    switchPorts:
      - portId: '1'
        serial: Q234-ABCD-0001
    switchProfilePorts:
      - portId: '2'
        profile: '1234'
- name: Delete by id
  cisco.meraki.networks_switch_link_aggregations:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: absent
    linkAggregationId: string
    networkId: string
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "id": "string",
      "switchPorts": [
        {
          "serial": "string",
          "portId": "string"
        }
      ]
    }
"""



#!/usr/bin/python
# Copyright (c) 2024 Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = """
module: acme_certificate_order_info
author: Felix Fontein (@felixfontein)
version_added: 2.24.0
short_description: Obtain information for an ACME v2 order
description:
  - Obtain information for an ACME v2 order.
    This can be used during the process of obtaining a new certificate with the
    L(ACME protocol,https://tools.ietf.org/html/rfc8555) from a Certificate
    Authority such as L(Let's Encrypt,https://letsencrypt.org/).
    This module does not support ACME v1, the original version of the ACME protocol before standardization.
  - This module needs to be used in conjunction with the
    M(community.crypto.acme_certificate_order_create),
    M(community.crypto.acme_certificate_order_validate), and
    M(community.crypto.acme_certificate_order_finalize) modules.
  - The module supports the V(dns-account-01) challenge type according to
    L(acme-dns-account-label draft 02, https://datatracker.ietf.org/doc/html/draft-ietf-acme-dns-account-label-02).
    Note that the supported draft version can change at any time,
    and changes will only be considered breaking once the draft reached RFC status.
  - The module supports the V(dns-persist-01) challenge type according to
    L(acme-dns-persist draft 01, https://www.ietf.org/archive/id/draft-ietf-acme-dns-persist-01.html).
    Note that the supported draft version can change at any time,
    and changes will only be considered breaking once the draft reached RFC status.
seealso:
  - module: community.crypto.acme_certificate_order_create
    description: Create an ACME order.
  - module: community.crypto.acme_certificate_order_validate
    description: Validate pending authorizations of an ACME order.
  - module: community.crypto.acme_certificate_order_finalize
    description: Finalize an ACME order after satisfying the challenges.
  - name: Automatic Certificate Management Environment (ACME)
    description: The specification of the ACME protocol (RFC 8555).
    link: https://tools.ietf.org/html/rfc8555
  - name: ACME TLS ALPN Challenge Extension
    description: The specification of the V(tls-alpn-01) challenge (RFC 8737).
    link: https://www.rfc-editor.org/rfc/rfc8737.html
  - module: community.crypto.acme_inspect
    description: Allows to debug problems.
  - module: community.crypto.acme_certificate_deactivate_authz
    description: Allows to deactivate (invalidate) ACME v2 orders.
extends_documentation_fragment:
  - community.crypto._acme.basic
  - community.crypto._acme.account
  - community.crypto._attributes
  - community.crypto._attributes.actiongroup_acme
  - community.crypto._attributes.idempotent_not_modify_state
  - community.crypto._attributes.info_module
options:
  order_uri:
    description:
      - The order URI provided by RV(community.crypto.acme_certificate_order_create#module:order_uri).
    type: str
    required: true
"""

EXAMPLES = r"""
---
- name: Create a challenge for sample.com using a account key from a variable
  community.crypto.acme_certificate_order_create:
    account_key_content: "{{ account_private_key }}"
    csr: /etc/pki/cert/csr/sample.com.csr
  register: order

- name: Obtain information on the order
  community.crypto.acme_certificate_order_info:
    account_key_src: /etc/pki/cert/private/account.key
    order_uri: "{{ order.order_uri }}"
  register: order_info

- name: Show information
  ansible.builtin.debug:
    var: order_info
"""

RETURN = """
account_uri:
  description: ACME account URI.
  returned: success
  type: str
order_uri:
  description: ACME order URI.
  returned: success
  type: str
order:
  description:
    - The order object.
    - See U(https://www.rfc-editor.org/rfc/rfc8555#section-7.1.3) for its specification.
  returned: success
  type: dict
  contains:
    status:
      description:
        - The status of this order.
        - See U(https://www.rfc-editor.org/rfc/rfc8555#section-7.1.6) for state changes.
      type: str
      returned: always
      choices:
        - pending
        - ready
        - processing
        - valid
        - invalid
    expires:
      description:
        - The timestamp after which the server will consider this order invalid.
        - Encoded in the format specified in L(RFC 3339, https://www.rfc-editor.org/rfc/rfc3339).
      type: str
      returned: if RV(order.status) is V(pending) or V(valid), and sometimes in other situations
    identifiers:
      description:
        - An array of identifier objects that the order pertains to.
      returned: always
      type: list
      elements: dict
      contains:
        type:
          description:
            - The type of identifier.
            - So far V(dns) and V(ip) are defined values.
          type: str
          returned: always
          sample: dns
          choices:
            - dns
            - ip
        value:
          description:
            - The identifier itself.
          type: str
          returned: always
          sample: example.com
    notBefore:
      description:
        - The requested value of the C(notBefore) field in the certificate.
        - Encoded in the date format defined in L(RFC 3339, https://www.rfc-editor.org/rfc/rfc3339).
      type: str
      returned: depending on order
    notAfter:
      description:
        - The requested value of the C(notAfter) field in the certificate.
        - Encoded in the date format defined in L(RFC 3339, https://www.rfc-editor.org/rfc/rfc3339).
      type: str
      returned: depending on order
    error:
      description:
        - The error that occurred while processing the order, if any.
        - This field is structured as a L(problem document according to RFC 7807, https://www.rfc-editor.org/rfc/rfc7807).
      type: dict
      returned: sometimes
    authorizations:
      description:
        - For pending orders, the authorizations that the client needs to complete before the
          requested certificate can be issued, including unexpired authorizations that the client
          has completed in the past for identifiers specified in the order.
        - The authorizations required are dictated by server policy; there may not be a 1:1
          relationship between the order identifiers and the authorizations required.
        - For final orders (in the V(valid) or V(invalid) state), the authorizations that were
          completed.  Each entry is a URL from which an authorization can be fetched with a POST-as-GET request.
        - The authorizations themselves are returned as RV(authorizations_by_identifier).
      type: list
      elements: str
      returned: always
    finalize:
      description:
        - A URL that a CSR must be POSTed to once all of the order's authorizations are satisfied to finalize the
          order.  The result of a successful finalization will be the population of the certificate URL for the order.
      type: str
      returned: always
    certificate:
      description:
        - A URL for the certificate that has been issued in response to this order.
      type: str
      returned: when the certificate has been issued
    replaces:
      description:
        - If the order was created to replace an existing certificate using the C(replaces) mechanism from
          L(RFC 9773, https://www.rfc-editor.org/rfc/rfc9773.html), this provides the
          certificate ID of the certificate that will be replaced by this order.
      type: str
      returned: when the certificate order is replacing a certificate through RFC 9773
    profile:
      description:
        - If the ACME CA supports profiles through the L(draft-aaron-acme-profiles,
          https://datatracker.ietf.org/doc/draft-aaron-acme-profiles/) mechanism and informs about the profile
          selected for this order, this field will contain the name of the profile used.
      type: str
      returned: depending on the ACME CA
authorizations_by_identifier:
  description:
    - A dictionary mapping identifiers to their authorization objects.
  returned: success
  type: dict
  contains:
    identifier:
      description:
        - The keys in this dictionary are the identifiers. C(identifier) is a placeholder used in the documentation.
        - See U(https://www.rfc-editor.org/rfc/rfc8555#section-7.1.4) for how authorization objects look like.
      type: dict
      contains:
        identifier:
          description:
            - The identifier that the account is authorized to represent.
          type: dict
          returned: always
          contains:
            type:
              description:
                - The type of identifier.
                - So far V(dns) and V(ip) are defined values.
              type: str
              returned: always
              sample: dns
              choices:
                - dns
                - ip
            value:
              description:
                - The identifier itself.
              type: str
              returned: always
              sample: example.com
        status:
          description:
            - The status of this authorization.
            - See U(https://www.rfc-editor.org/rfc/rfc8555#section-7.1.6) for state changes.
          type: str
          choices:
            - pending
            - valid
            - invalid
            - deactivated
            - expired
            - revoked
          returned: always
        expires:
          description:
            - The timestamp after which the server will consider this authorization invalid.
            - Encoded in the format specified in L(RFC 3339, https://www.rfc-editor.org/rfc/rfc3339).
          type: str
          returned: if RV(authorizations_by_identifier.identifier.status=valid), and sometimes in other situations
        challenges:
          description:
            - For pending authorizations, the challenges that the client can fulfill in order to prove
              possession of the identifier.
            - For valid authorizations, the challenge that was validated.
            - For invalid authorizations, the challenge that was attempted and failed.
            - Each array entry is an object with parameters required to validate the challenge.
              A client should attempt to fulfill one of these challenges, and a server should consider
              any one of the challenges sufficient to make the authorization valid.
            - See U(https://www.rfc-editor.org/rfc/rfc8555#section-8) for the general structure. The structure
              of every entry depends on the challenge's type. For C(tls-alpn-01) challenges, the structure is
              defined in U(https://www.rfc-editor.org/rfc/rfc8737.html#section-3).
          type: list
          elements: dict
          returned: always
          contains:
            type:
              description:
                - The type of challenge encoded in the object.
                - Support for V(dns-account-01) and V(dns-persist-01) has been added in community.crypto 3.2.0.
              type: str
              returned: always
              choices:
                - http-01
                - dns-01
                - dns-account-01
                - dns-persist-01
                - tls-alpn-01
            url:
              description:
                - The URL to which a response can be posted.
              type: str
              returned: always
            status:
              description:
                - The status of this challenge.
                - See U(https://www.rfc-editor.org/rfc/rfc8555#section-7.1.6) for state changes.
              type: str
              choices:
                - pending
                - processing
                - valid
                - invalid
              returned: always
            validated:
              description:
                - The time at which the server validated this challenge.
                - Encoded in the format specified in L(RFC 3339, https://www.rfc-editor.org/rfc/rfc3339).
              type: str
              returned: always if RV(authorizations_by_identifier.identifier.challenges[].type=valid), otherwise in some situations
            error:
              description:
                - Error that occurred while the server was validating the challenge, if any.
                - This field is structured as a L(problem document according to RFC 7807, https://www.rfc-editor.org/rfc/rfc7807).
              type: dict
              returned: always if RV(authorizations_by_identifier.identifier.challenges[].type=invalid), otherwise in some situations
        wildcard:
          description:
            - This field B(must) be present and true for authorizations created as a result of a
              C(newOrder) request containing a DNS identifier with a value that was a wildcard
              domain name.  For other authorizations, it B(must) be absent.
            - Wildcard domain names are described in U(https://www.rfc-editor.org/rfc/rfc8555#section-7.1.3)
              of the ACME specification.
          type: bool
          returned: sometimes
authorizations_by_status:
  description:
    - For every status, a list of identifiers whose authorizations have this status.
  returned: success
  type: dict
  contains:
    pending:
      description:
        - A list of all identifiers whose authorizations are in the C(pending) state.
        - See U(https://www.rfc-editor.org/rfc/rfc8555#section-7.1.6) for state changes
          of authorizations.
      type: list
      elements: str
      returned: always
    invalid:
      description:
        - A list of all identifiers whose authorizations are in the C(invalid) state.
        - See U(https://www.rfc-editor.org/rfc/rfc8555#section-7.1.6) for state changes
          of authorizations.
      type: list
      elements: str
      returned: always
    valid:
      description:
        - A list of all identifiers whose authorizations are in the C(valid) state.
        - See U(https://www.rfc-editor.org/rfc/rfc8555#section-7.1.6) for state changes
          of authorizations.
      type: list
      elements: str
      returned: always
    revoked:
      description:
        - A list of all identifiers whose authorizations are in the C(revoked) state.
        - See U(https://www.rfc-editor.org/rfc/rfc8555#section-7.1.6) for state changes
          of authorizations.
      type: list
      elements: str
      returned: always
    deactivated:
      description:
        - A list of all identifiers whose authorizations are in the C(deactivated) state.
        - See U(https://www.rfc-editor.org/rfc/rfc8555#section-7.1.6) for state changes
          of authorizations.
      type: list
      elements: str
      returned: always
    expired:
      description:
        - A list of all identifiers whose authorizations are in the C(expired) state.
        - See U(https://www.rfc-editor.org/rfc/rfc8555#section-7.1.6) for state changes
          of authorizations.
      type: list
      elements: str
      returned: always
"""

import typing as t

from ansible_collections.community.crypto.plugins.module_utils._acme.acme import (
    create_backend,
    create_default_argspec,
)
from ansible_collections.community.crypto.plugins.module_utils._acme.certificate import (
    ACMECertificateClient,
)
from ansible_collections.community.crypto.plugins.module_utils._acme.errors import (
    ModuleFailException,
)


def main() -> t.NoReturn:
    argument_spec = create_default_argspec(with_certificate=False)
    argument_spec.update_argspec(
        order_uri={"type": "str", "required": True},
    )
    module = argument_spec.create_ansible_module(supports_check_mode=True)

    backend = create_backend(module, needs_acme_v2=False)

    try:
        client = ACMECertificateClient(module=module, backend=backend)
        order = client.load_order()
        authorizations_by_identifier: dict[str, dict[str, t.Any]] = {}
        authorizations_by_status: dict[str, list[str]] = {
            "pending": [],
            "invalid": [],
            "valid": [],
            "revoked": [],
            "deactivated": [],
            "expired": [],
        }
        for identifier, authz in order.authorizations.items():
            authorizations_by_identifier[identifier] = authz.to_json()
            if authz.status is not None:
                authorizations_by_status[authz.status].append(identifier)
        module.exit_json(
            changed=False,
            account_uri=client.client.account_uri,
            order_uri=order.url,
            order=order.data,
            authorizations_by_identifier=authorizations_by_identifier,
            authorizations_by_status=authorizations_by_status,
        )
    except ModuleFailException as e:
        e.do_fail(module=module)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type


class ModuleDocFragment(object):
    DOCUMENTATION = r"""
options:
  vrf:
    description:
    - The VRF reference details.
    type: dict
    suboptions:
      uuid:
        description:
        - The UUID of the VRF.
        - This parameter can be used instead of O(vrf.reference).
        type: str
      reference:
        description:
        - The reference details of the VRF.
        - This parameter can be used instead of O(vrf.uuid).
        type: dict
        aliases: [ ref ]
        suboptions:
          name:
            description:
            - The name of the VRF.
            required: true
            type: str
          schema:
            description:
            - The name of the schema associated with the VRF.
            - This parameter can be used instead of O(vrf.reference.schema_id).
            type: str
          schema_id:
            description:
            - The schema ID associated with the VRF.
            - This parameter can be used instead of O(vrf.reference.schema).
            type: str
          template:
            description:
            - The name of the template associated with the VRF.
            - This parameter can be used instead of O(vrf.reference.template_id).
            type: str
          template_id:
            description:
            - The template ID associated with the VRF.
            - This parameter can be used instead of O(vrf.reference.template).
            type: str
"""



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2025, Sabari Jaganathan (@sajagana) <sajagana@cisco.com>

# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "community"}

DOCUMENTATION = r"""
module: ndo_l3out_bgp_peer
version_added: "2.11.0"
short_description: Manage L3Out BGP Peer on Cisco Nexus Dashboard Orchestrator (NDO).
description:
- Manage L3Out BGP Peer on Cisco Nexus Dashboard Orchestrator (NDO).
- This module is only supported on ND v3.1 (NDO v4.3) and later.
author:
- Sabari Jaganathan (@sajagana)
options:
  template:
    description:
    - The name of the L3Out template.
    - This parameter or O(template_id) is required.
    type: str
    aliases: [ l3out_template ]
  template_id:
    description:
    - The ID of the L3Out template.
    - This parameter or O(template) is required.
    type: str
    aliases: [ l3out_template_id ]
  l3out:
    description:
    - The name of the L3Out.
    - This parameter or O(l3out_uuid) is required.
    type: str
  l3out_uuid:
    description:
    - The UUID of the L3Out.
    - This parameter or O(l3out) is required.
    type: str
  parent_type:
    description:
    - The parent type of the L3Out.
    type: str
    default: node_group
    choices: [ floating_svi, node_group, routed, routed_sub, svi ]
  node_group:
    description:
    - The name of the Node Group Policy.
    type: str
  node_id:
    description:
    - The ID of the node (border leaf switch).
    type: str
    aliases: [ node, border_leaf, anchor_node ]
  path:
    description:
    - The path of the interface.
    - The path must be an existing physical port (eg. eth1/1).
    type: str
    aliases: [ interface ]
  port_channel:
    description:
    - The port channel details.
    type: dict
    aliases: [ pc ]
    suboptions:
      uuid:
        description:
        - The UUID of the port channel.
        - This parameter or O(port_channel.reference) is required.
        type: str
      reference:
        description:
        - The reference details of the port channel.
        - This parameter or O(port_channel.uuid) is required.
        type: dict
        aliases: [ ref ]
        suboptions:
          name:
            description:
            - The name of the port channel.
            type: str
            required: true
          template:
            description:
            - The name of the template that contains the port channel.
            - This parameter or O(port_channel.reference.template_id) is required.
            type: str
          template_id:
            description:
            - The ID of the template that contains the port channel.
            - This parameter or O(port_channel.reference.template) is required.
            type: str
  virtual_port_channel:
    description:
    - The virtual port channel details.
    type: dict
    aliases: [ vpc ]
    suboptions:
      uuid:
        description:
        - The UUID of the virtual port channel.
        - This parameter or O(virtual_port_channel.reference) is required.
        type: str
      reference:
        description:
        - The reference details of the virtual port channel.
        - This parameter or O(virtual_port_channel.uuid) is required.
        type: dict
        aliases: [ ref ]
        suboptions:
          name:
            description:
            - The name of the virtual port channel.
            type: str
            required: true
          template:
            description:
            - The name of the template that contains the virtual port channel.
            - This parameter or O(virtual_port_channel.reference.template_id) is required.
            type: str
          template_id:
            description:
            - The ID of the template that contains the virtual port channel.
            - This parameter or O(virtual_port_channel.reference.template) is required.
            type: str
  encapsulation_type:
    description:
    - The encapsulation type of the interface.
    type: str
    aliases: [ encap_type ]
    choices: [ vlan, vxlan ]
  encapsulation_value:
    description:
    - The encapsulation value of the interface.
    - The option O(encapsulation_type=vlan), specifies VLAN ID which must be in the range 1 - 4094.
    - The option O(encapsulation_type=vxlan), specifies VXLAN Network Identifier (VNI) which must be in the range 5000 - 16777215.
    type: int
    aliases: [ encap, encapsulation, encapsulation_id ]
  ipv4_address:
    description:
    - The IPv4 address of the L3Out BGP Peer.
    - Providing an empty string will remove the O(ipv4_address="") from the L3Out BGP Peer.
    type: str
    aliases: [ peer_address_ipv4 ]
  ipv6_address:
    description:
    - The IPv6 address of the L3Out BGP Peer.
    - Providing an empty string will remove the O(ipv6_address="") from the L3Out BGP Peer.
    type: str
    aliases: [ peer_address_ipv6 ]
  remote_asn:
    description:
    - The remote autonomous system number (ASN) of the L3Out BGP Peer.
    - The value must be between 1 and 4294967295.
    - Providing an empty string will remove the O(remote_asn="") from the L3Out BGP Peer.
    type: str
  admin_state:
    description:
    - The administrative state of the L3Out BGP Peer.
    - Defaults to C(enabled) when unset during creation.
    type: str
    choices: [ enabled, disabled ]
  auth_password:
    description:
    - The authentication password of the L3Out BGP Peer.
    type: str
  ebgp_multi_hop_ttl:
    description:
    - The TTL for eBGP multi-hop of the L3Out BGP Peer.
    - Defaults to 1 when unset during creation.
    - The value must be between 1 and 255.
    type: int
  site_of_origin:
    description:
    - The site of origin for the L3Out BGP Peer. The value must adhere to the pattern "extended:as2-nn2:1000:65534".
    type: str
    aliases: [ fabric_of_origin ]
  weight:
    description:
    - The weight of the L3Out BGP Peer.
    - The value must be between 1 and 65535.
    - Providing an empty string will remove the O(weight="") from the L3Out BGP Peer.
    type: str
  allowed_self_as_count:
    description:
    - The allowed self-AS count of the L3Out BGP Peer.
    - Defaults to 3 when unset during creation.
    - The value must be between 1 and 10.
    type: int
  local_asn_config:
    description:
    - The configuration for local ASN of the L3Out BGP Peer.
    type: str
    choices: [ none, no_prepend, dual_as, replace_as ]
  local_asn:
    description:
    - The local autonomous system number (ASN) of the L3Out BGP Peer.
    - The value must be between 1 and 4294967295.
    - Providing an empty string will remove the O(local_asn="") from the L3Out BGP Peer.
    type: str
  import_route_map:
    description:
    - The name of the import route map.
    - Providing an empty dictionary will remove the O(import_route_map={}) from the L3Out BGP Peer.
    - This parameter can be used instead of O(import_route_map_uuid).
    type: dict
    suboptions:
      name:
        description:
        - The name of the import route map.
        type: str
      template:
        description:
        - The template associated with the import route map.
        - This parameter or O(import_route_map.template_id) is required.
        type: str
      template_id:
        description:
        - The template ID associated with the the import route map.
        - This parameter or O(import_route_map.template) is required.
        type: str
  import_route_map_uuid:
    description:
    - The UUID of the import route map.
    - Providing an empty string will remove the O(import_route_map_uuid="") from the L3Out BGP Peer.
    - This parameter can be used instead of O(import_route_map).
    type: str
  export_route_map:
    description:
    - The name of the export route map.
    - Providing an empty dictionary will remove the O(export_route_map={}) from the L3Out BGP Peer.
    - This parameter can be used instead of O(export_route_map_uuid).
    type: dict
    suboptions:
      name:
        description:
        - The name of the export route map.
        type: str
      template:
        description:
        - The template associated with the export route map.
        - This parameter or O(export_route_map.template_id) is required.
        type: str
      template_id:
        description:
        - The template ID associated with the the export route map.
        - This parameter or O(export_route_map.template) is required.
        type: str
  export_route_map_uuid:
    description:
    - The UUID of the export route map.
    - Providing an empty string will remove the O(export_route_map_uuid="") from the L3Out BGP Peer.
    - This parameter can be used instead of O(export_route_map).
    type: str
  peer_prefix:
    description:
    - The name of the peer prefix.
    - Providing an empty dictionary will remove the O(peer_prefix={}) from the L3Out BGP Peer.
    - This parameter can be used instead of O(peer_prefix_uuid).
    type: dict
    suboptions:
      name:
        description:
        - The name of the peer prefix.
        type: str
      template:
        description:
        - The template associated with the peer prefix.
        - This parameter or O(peer_prefix.template_id) is required.
        type: str
      template_id:
        description:
        - The template ID associated with the the peer prefix.
        - This parameter or O(peer_prefix.template) is required.
        type: str
  peer_prefix_uuid:
    description:
    - The UUID of the peer prefix.
    - Providing an empty string will remove the O(peer_prefix_uuid="") from the L3Out BGP Peer.
    - This parameter can be used instead of O(peer_prefix).
    type: str
  bgp_controls:
    description:
    - The BGP control settings for the peer.
    type: dict
    suboptions:
      state:
        description:
        - Use C(enabled) to configure the BGP control settings.
        - Use C(disabled) to remove the BGP control settings.
        type: str
        choices: [ enabled, disabled ]
      allow_self_as:
        description:
        - The allow self AS flag of the BGP control.
        type: bool
      override_as:
        description:
        - The override AS flag of the BGP control.
        type: bool
      disabled_peer_as_check:
        description:
        - The disable peer AS check flag of the BGP control.
        type: bool
      next_hop_self:
        description:
        - The set next hop to self flag of the BGP control.
        type: bool
      send_community:
        description:
        - The send community flag of the BGP control.
        type: bool
      send_extended_community:
        description:
        - The send extended community flag of the BGP control.
        type: bool
      send_domain_path:
        description:
        - The send domain path flag of the BGP control.
        type: bool
  peer_controls:
    description:
    - The Peer control settings for the BGP peer.
    type: dict
    suboptions:
      state:
        description:
        - Use C(enabled) to configure the Peer control settings.
        - Use C(disabled) to remove the Peer control settings.
        type: str
        choices: [ enabled, disabled ]
      bfd:
        description:
        - The Enable Bidirectional Forwarding Detection (BFD) flag of the Peer control.
        type: bool
      disable_peer_connected_check:
        description:
        - The disable peer connected check flag of the Peer control.
        type: bool
  address_families:
    description:
    - The address family controls for the BGP peer.
    type: dict
    suboptions:
      state:
        description:
        - Use C(enabled) to configure the address control settings.
        - Use C(disabled) to remove the address control settings.
        type: str
        choices: [ enabled, disabled ]
      multicast:
        description:
        - The multicast address flag of the address family.
        type: bool
      unicast:
        description:
        - The unicast address flag of the address family.
        type: bool
  private_as_controls:
    description:
    - The private AS control settings for the BGP peer.
    type: dict
    suboptions:
      state:
        description:
        - Use C(enabled) to configure the private AS control settings.
        - Use C(disabled) to remove the private AS control settings.
        type: str
        choices: [ enabled, disabled ]
      remove_all:
        description:
        - The remove all flag of the private AS numbers.
        type: bool
      replace_with_local_as:
        description:
        - The replace private AS with local AS flag of the private AS control.
        type: bool
  state:
    description:
    - Use C(absent) for removing.
    - Use C(query) for listing an object or multiple objects.
    - Use C(present) for creating or updating.
    type: str
    choices: [ absent, query, present ]
    default: query
notes:
- The O(template) must exist before using this module in your playbook.
  Use M(cisco.mso.ndo_template) to create the L3Out template.
- The O(l3out) must exist before using this module in your playbook.
  Use M(cisco.mso.ndo_l3out_template) to create the L3Out.
- The O(node_group) must exist before using this module in your playbook.
  Use M(cisco.mso.ndo_l3out_node_group_policy) to create the L3Out Node Group Policy.
- The O(peer_prefix) must exist before using it with this module in your playbook.
  Use M(cisco.mso.ndo_tenant_bgp_peer_prefix_policy) to create the BGP Peer Prefix Policy.
seealso:
- module: cisco.mso.ndo_template
- module: cisco.mso.ndo_l3out_template
- module: cisco.mso.ndo_l3out_node_group_policy
- module: cisco.mso.ndo_tenant_bgp_peer_prefix_policy
extends_documentation_fragment: cisco.mso.modules
"""

EXAMPLES = r"""
- name: Create an L3Out BGP Peer with minimum configuration
  cisco.mso.ndo_l3out_bgp_peer:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: l3out_template
    l3out: l3out_1
    node_group: node_group_policy_1
    ipv4_address: "1.1.1.1"
    ipv6_address: "1::8/16"
    auth_password: 123
    state: present

- name: Create L3Out BGP Peer on a L3out routed interface of type port
  cisco.mso.ndo_l3out_bgp_peer:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: l3out_template
    l3out: l3out_1
    ipv4_address: 5.5.5.5
    parent_type: routed
    node_id: 101
    path: eth1/1
    state: present

- name: Create a L3out routed sub-interface of type port channel
  cisco.mso.ndo_l3out_routed_sub_interface:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: l3out_template
    l3out: l3out_1
    ipv4_address: 5.5.5.5
    parent_type: routed_sub
    port_channel:
      uuid: '{{ port_channel_interface_1.current.uuid }}'
    encapsulation_type: vlan
    encapsulation_value: 100
    state: present

- name: Create L3Out BGP Peer on a L3out svi interface of type virtual port channel
  cisco.mso.ndo_l3out_bgp_peer:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: l3out_template
    l3out: l3out_1
    ipv4_address: 5.5.5.5
    parent_type: svi
    virtual_port_channel:
      uuid: '{{ virtual_port_channel_interface_1.current.uuid }}'
    encapsulation_type: vlan
    encapsulation_value: 100
    state: present

- name: Create L3Out BGP Peer on a L3out floating svi interface
  cisco.mso.ndo_l3out_bgp_peer:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: l3out_template
    l3out: l3out_1
    ipv4_address: "1.1.1.1"
    parent_type: floating_svi
    node_id: 101
    encapsulation_type: vlan
    encapsulation_value: 100
    state: present

- name: Update an L3Out BGP Peer with full configuration
  cisco.mso.ndo_l3out_bgp_peer:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: l3out_template
    l3out: l3out_1
    node_group: node_group_policy_1
    ipv4_address: "1.1.1.1"
    ipv6_address: "1::8/16"
    remote_asn: 2
    admin_state: enabled
    import_route_map:
      name: ans_route_map
      template: ansible_test_policy
    export_route_map:
      name: ans_route_map_2
      template: ansible_test_policy
    peer_prefix:
      name: ansible_test_bgp_peer_prefix_policy
      template: ansible_test_policy
    ebgp_multi_hop_ttl: 1
    auth_password: 123
    weight: 2
    site_of_origin: "extended:as2-nn2:1000:65534"
    allowed_self_as_count: 3
    local_asn_config: replace_as
    local_asn: 1
    bgp_controls:
      allow_self_as: true
      override_as: true
      disabled_peer_as_check: true
      next_hop_self: true
      send_community: true
      send_extended_community: true
      send_domain_path: true
    peer_controls:
      bfd: true
      disable_peer_connected_check: true
    address_families:
      multicast: true
      unicast: true
    private_as_controls:
      remove_all: true
      replace_with_local_as: true
    state: present

- name: Query an L3Out BGP Peer with IPv4 and IPv6 addresses
  cisco.mso.ndo_l3out_bgp_peer:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: l3out_template
    l3out: l3out_1
    node_group: node_group_policy_1
    ipv4_address: "1.1.1.1"
    ipv6_address: "1::8/16"
    state: query
  register: query_with_ipv4_and_ipv6

- name: Query an L3Out BGP Peer with IPv4
  cisco.mso.ndo_l3out_bgp_peer:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: l3out_template
    l3out: l3out_1
    node_group: node_group_policy_1
    ipv4_address: "1.1.1.1"
    state: query
  register: query_with_ipv4

- name: Query an L3Out BGP Peer with IPv6
  cisco.mso.ndo_l3out_bgp_peer:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: l3out_template
    l3out: l3out_1
    node_group: node_group_policy_1
    ipv6_address: "1::8/16"
    state: query
  register: query_with_ipv6

- name: Query all L3Out BGP Peer
  cisco.mso.ndo_l3out_bgp_peer:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: l3out_template
    l3out: l3out_1
    node_group: node_group_policy_1
    state: query
  register: query_all

- name: Remove an L3Out BGP Peer with IPv4 and IPv6 addresses
  cisco.mso.ndo_l3out_bgp_peer:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: l3out_template
    l3out: l3out_1
    node_group: node_group_policy_1
    ipv4_address: "1.1.1.1"
    ipv6_address: "1::8/16"
    state: absent

- name: Remove an L3Out BGP Peer with IPv4
  cisco.mso.ndo_l3out_bgp_peer:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: l3out_template
    l3out: l3out_1
    node_group: node_group_policy_1
    ipv4_address: "1.1.1.1"
    state: absent

- name: Remove an L3Out BGP Peer with IPv6
  cisco.mso.ndo_l3out_bgp_peer:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: l3out_template
    l3out: l3out_1
    node_group: node_group_policy_1
    ipv6_address: "1::8/16"
    state: absent
"""

RETURN = r"""
"""


import copy
from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.mso.plugins.module_utils.mso import (
    MSOModule,
    mso_argument_spec,
    ndo_template_object_spec,
    ndo_l3out_port_channel_spec,
    ndo_l3out_virtual_port_channel_spec,
)
from ansible_collections.cisco.mso.plugins.module_utils.template import KVPair
from ansible_collections.cisco.mso.plugins.module_utils.templates import MSOTemplates
from ansible_collections.cisco.mso.plugins.module_utils.constants import LOCAL_ASN_CONFIG
from ansible_collections.cisco.mso.plugins.module_utils.utils import append_update_ops_data, check_if_all_elements_are_none


def main():
    argument_spec = mso_argument_spec()
    argument_spec.update(
        template=dict(type="str", aliases=["l3out_template"]),
        template_id=dict(type="str", aliases=["l3out_template_id"]),
        l3out=dict(type="str"),
        l3out_uuid=dict(type="str"),
        # Default is use because the first release of the module only supported node_group type
        # This way no changes are needed for existing playbooks
        parent_type=dict(type="str", choices=["floating_svi", "node_group", "routed", "routed_sub", "svi"], default="node_group"),
        node_group=dict(type="str"),
        node_id=dict(type="str", aliases=["node", "border_leaf", "anchor_node"]),
        path=dict(type="str", aliases=["interface"]),
        port_channel=ndo_l3out_port_channel_spec(micro_bfd=False),
        virtual_port_channel=ndo_l3out_virtual_port_channel_spec(side_b=False),
        encapsulation_type=dict(type="str", choices=["vlan", "vxlan"], aliases=["encap_type"]),
        encapsulation_value=dict(type="int", aliases=["encap", "encapsulation", "encapsulation_id"]),
        ipv4_address=dict(type="str", aliases=["peer_address_ipv4"]),
        ipv6_address=dict(type="str", aliases=["peer_address_ipv6"]),
        remote_asn=dict(type="str"),
        admin_state=dict(type="str", choices=["enabled", "disabled"]),
        ebgp_multi_hop_ttl=dict(type="int"),
        auth_password=dict(type="str", no_log=True),
        weight=dict(type="str"),
        site_of_origin=dict(type="str", aliases=["fabric_of_origin"]),
        allowed_self_as_count=dict(type="int"),
        local_asn_config=dict(type="str", choices=list(LOCAL_ASN_CONFIG)),
        local_asn=dict(type="str"),
        bgp_controls=dict(
            type="dict",
            options=dict(
                state=dict(type="str", choices=["enabled", "disabled"]),
                allow_self_as=dict(type="bool"),
                override_as=dict(type="bool"),
                disabled_peer_as_check=dict(type="bool"),
                next_hop_self=dict(type="bool"),
                send_community=dict(type="bool"),
                send_extended_community=dict(type="bool"),
                send_domain_path=dict(type="bool"),
            ),
        ),
        peer_controls=dict(
            type="dict",
            options=dict(
                state=dict(type="str", choices=["enabled", "disabled"]),
                bfd=dict(type="bool"),
                disable_peer_connected_check=dict(type="bool"),
            ),
        ),
        address_families=dict(
            type="dict",
            options=dict(
                state=dict(type="str", choices=["enabled", "disabled"]),
                multicast=dict(type="bool"),
                unicast=dict(type="bool"),
            ),
        ),
        private_as_controls=dict(
            type="dict",
            options=dict(
                state=dict(type="str", choices=["enabled", "disabled"]),
                remove_all=dict(type="bool"),
                replace_with_local_as=dict(type="bool"),
            ),
        ),
        import_route_map=ndo_template_object_spec(),
        import_route_map_uuid=dict(type="str"),
        export_route_map=ndo_template_object_spec(),
        export_route_map_uuid=dict(type="str"),
        peer_prefix=ndo_template_object_spec(),
        peer_prefix_uuid=dict(type="str"),
        state=dict(type="str", default="query", choices=["absent", "query", "present"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "present", ["ipv4_address", "ipv6_address"], True],
            ["state", "absent", ["ipv4_address", "ipv6_address"], True],
            ["parent_type", "node_group", ["node_group"]],
            ["parent_type", "routed", ["path", "port_channel"], True],
            ["parent_type", "routed_sub", ["path", "port_channel"], True],
            ["parent_type", "routed_sub", ["encapsulation_type", "encapsulation_value"]],
            ["parent_type", "svi", ["path", "port_channel", "virtual_port_channel"], True],
            ["parent_type", "svi", ["encapsulation_type", "encapsulation_value"]],
            ["parent_type", "floating_svi", ["node_id", "encapsulation_type", "encapsulation_value"]],
        ],
        required_by={"path": ("node_id")},
        mutually_exclusive=[
            ("template", "template_id"),
            ("l3out", "l3out_uuid"),
            ("import_route_map", "import_route_map_uuid"),
            ("export_route_map", "export_route_map_uuid"),
            ("peer_prefix", "peer_prefix_uuid"),
            ("port_channel", "path"),
            ("port_channel", "node_id"),
            ("port_channel", "virtual_port_channel"),
            ("virtual_port_channel", "path"),
            ("virtual_port_channel", "node_id"),
        ],
        required_one_of=[
            ["template", "template_id"],
            ["l3out", "l3out_uuid"],
        ],
    )

    mso = MSOModule(module)
    mso_templates = MSOTemplates(mso)

    template = module.params.get("template")
    template_id = module.params.get("template_id")
    l3out = module.params.get("l3out")
    l3out_uuid = module.params.get("l3out_uuid")
    parent_type = module.params.get("parent_type")
    ipv4_addr = module.params.get("ipv4_address")
    ipv6_addr = module.params.get("ipv6_address")
    remote_asn = module.params.get("remote_asn")
    admin_state = module.params.get("admin_state")
    auth_password = module.params.get("auth_password")
    ebgp_multi_hop_ttl = module.params.get("ebgp_multi_hop_ttl")
    site_of_origin = module.params.get("site_of_origin")
    weight = module.params.get("weight")
    allowed_self_as_count = module.params.get("allowed_self_as_count")
    local_asn_config = LOCAL_ASN_CONFIG.get(module.params.get("local_asn_config"))
    local_asn = module.params.get("local_asn")

    import_route_map = module.params.get("import_route_map")
    import_route_map_uuid = module.params.get("import_route_map_uuid")
    export_route_map = module.params.get("export_route_map")
    export_route_map_uuid = module.params.get("export_route_map_uuid")
    peer_prefix = module.params.get("peer_prefix")
    peer_prefix_uuid = module.params.get("peer_prefix_uuid")

    bgp_controls = module.params.get("bgp_controls")
    peer_controls = module.params.get("peer_controls")
    address_families = module.params.get("address_families")
    private_as_controls = module.params.get("private_as_controls")
    state = module.params.get("state")

    mso_template = mso_templates.get_template("l3out", template, template_id)
    mso_template.validate_template("l3out")
    mso.check_template_when_name_is_provided(import_route_map)
    mso.check_template_when_name_is_provided(export_route_map)
    mso.check_template_when_name_is_provided(peer_prefix)

    l3out_object = mso_template.get_l3out_object(uuid=l3out_uuid, name=l3out, fail_module=True)
    parent_object, parent_path = mso_template.get_parent_details_for_nested_object_in_l3out(mso_templates, l3out_object)

    bgp_peer_objects = get_bgp_peer_by_address(mso_template, parent_object.details.get("bgpPeers", []), ipv4_addr=ipv4_addr, ipv6_addr=ipv6_addr)

    if bgp_peer_objects and (ipv4_addr or ipv6_addr):
        set_bgp_peer_details(mso_template, parent_type, parent_object.details, bgp_peer_objects.details)
        mso.existing = copy.deepcopy(bgp_peer_objects.details)  # Query a specific object
        mso.previous = copy.deepcopy(bgp_peer_objects.details)  # Query a specific object
    elif bgp_peer_objects:
        mso.existing = [set_bgp_peer_details(mso_template, parent_type, parent_object.details, bgp_peer) for bgp_peer in bgp_peer_objects]  # Query all objects

    if state != "query":
        bgp_peer_path = "{0}/bgpPeers/{1}".format(parent_path, bgp_peer_objects.index if bgp_peer_objects else "-")

    ops = []

    if state == "present":
        peer_prefix_is_empty = check_if_all_elements_are_none(peer_prefix.values()) if peer_prefix else True  # peer_prefix is None or empty dict => True
        import_route_map_is_empty = check_if_all_elements_are_none(import_route_map.values()) if import_route_map else True
        export_route_map_is_empty = check_if_all_elements_are_none(export_route_map.values()) if export_route_map else True

        mso_templates = MSOTemplates(mso)
        if not peer_prefix_is_empty:
            peer_prefix_template = mso_templates.get_template("tenant", peer_prefix.get("template"), peer_prefix.get("template_id"))
            peer_prefix_uuid = peer_prefix_template.get_template_policy_uuid("tenant", peer_prefix.get("name"), "bgpPeerPrefixPolicies")

        if not import_route_map_is_empty:
            import_route_map_template = mso_templates.get_template("tenant", import_route_map.get("template"), import_route_map.get("template_id"))
            import_route_map_uuid = import_route_map_template.get_template_policy_uuid("tenant", import_route_map.get("name"), "routeMapPolicies")

        if not export_route_map_is_empty:
            export_route_map_template = mso_templates.get_template("tenant", export_route_map.get("template"), export_route_map.get("template_id"))
            export_route_map_uuid = export_route_map_template.get_template_policy_uuid("tenant", export_route_map.get("name"), "routeMapPolicies")

        mso_values = dict(
            peerAddressV4=ipv4_addr,
            peerAddressV6=ipv6_addr,
            adminState=admin_state,
            authEnabled=True if auth_password else False,
            allowedSelfASCount=allowed_self_as_count,
            ebpgMultiHopTTL=ebgp_multi_hop_ttl,
            siteOfOrigin=site_of_origin,
            localAsnConfig=local_asn_config,
            peerPrefixRef=peer_prefix_uuid,
            importRouteMapRef=import_route_map_uuid,
            exportRouteMapRef=export_route_map_uuid,
            password=dict(value=auth_password) if auth_password is not None else None,
            peerAsn=int(remote_asn) if remote_asn not in ["", None] else None,
            weight=int(weight) if weight not in ["", None] else None,
            localAsn=int(local_asn) if local_asn not in ["", None] else None,
        )

        if not mso.existing:
            # BGP Controls
            if bgp_controls:
                bgp_controls_dict = dict()

                if bgp_controls["allow_self_as"]:
                    bgp_controls_dict["allowSelfAS"] = bgp_controls["allow_self_as"]

                if bgp_controls["override_as"]:
                    bgp_controls_dict["asOverride"] = bgp_controls["override_as"]

                if bgp_controls["disabled_peer_as_check"]:
                    bgp_controls_dict["disablePeerASCheck"] = bgp_controls["disabled_peer_as_check"]

                if bgp_controls["next_hop_self"]:
                    bgp_controls_dict["nextHopSelf"] = bgp_controls["next_hop_self"]

                if bgp_controls["send_community"]:
                    bgp_controls_dict["sendCommunity"] = bgp_controls["send_community"]

                if bgp_controls["send_extended_community"]:
                    bgp_controls_dict["sendDomainPath"] = bgp_controls["send_extended_community"]

                if bgp_controls["send_domain_path"]:
                    bgp_controls_dict["sendExtendedCommunity"] = bgp_controls["send_domain_path"]

                mso_values["bgpControls"] = bgp_controls_dict

            # Peer Controls
            if peer_controls:
                peer_controls_dict = dict()

                if peer_controls["bfd"]:
                    peer_controls_dict["bfd"] = peer_controls["bfd"]

                if peer_controls["disable_peer_connected_check"]:
                    peer_controls_dict["disableConnectedCheck"] = peer_controls["disable_peer_connected_check"]

                mso_values["peerControls"] = peer_controls_dict

            # Address Type Controls
            if address_families:
                address_families_dict = dict()

                if address_families["multicast"]:
                    address_families_dict["afMast"] = address_families["multicast"]

                if address_families["unicast"]:
                    address_families_dict["afUcast"] = address_families["unicast"]

                mso_values["addressTypeControls"] = address_families_dict

            # Private AS Controls
            if private_as_controls:
                private_as_controls_dict = dict()

                if private_as_controls["remove_all"]:
                    private_as_controls_dict["removeAll"] = private_as_controls["remove_all"]

                if private_as_controls["replace_with_local_as"]:
                    private_as_controls_dict["replaceWithLocalAS"] = private_as_controls["replace_with_local_as"]

                mso_values["privateASControls"] = private_as_controls_dict

            if private_as_controls and (private_as_controls.get("remove_all") or private_as_controls.get("replace_with_local_as")):
                mso_values["privateASControls"]["removeExclusive"] = True

            mso.sanitize(mso_values)
            ops.append(dict(op="add", path=bgp_peer_path, value=mso.sent))

        elif mso.existing:
            required_keys = ["peerPrefixRef", "importRouteMapRef", "exportRouteMapRef"]
            proposed_payload = copy.deepcopy(mso.existing)
            mso_values_remove = list()

            if ipv4_addr == "" and "peerAddressV4" in proposed_payload:
                mso_values_remove.append("peerAddressV4")
                mso_values.pop("peerAddressV4", None)

            if ipv6_addr == "" and "peerAddressV6" in proposed_payload:
                mso_values_remove.append("peerAddressV6")
                mso_values.pop("peerAddressV6", None)

            if remote_asn == "" and "peerAsn" in proposed_payload:
                mso_values_remove.append("peerAsn")
                mso_values.pop("peerAsn", None)

            if weight == "" and "weight" in proposed_payload:
                mso_values_remove.append("weight")
                mso_values.pop("weight", None)

            if local_asn == "" and "localAsn" in proposed_payload:
                mso_values_remove.append("localAsn")
                mso_values.pop("localAsn", None)

            if site_of_origin == "" and "siteOfOrigin" in proposed_payload:
                mso_values_remove.append("siteOfOrigin")
                mso_values.pop("siteOfOrigin", None)

            if mso.existing.get("password", {}).get("ref"):
                mso.existing["password"].pop("ref", None)
                mso.previous["password"].pop("ref", None)

            if auth_password == "" and "password" in proposed_payload:
                mso_values["authEnabled"] = False
                mso_values.pop("password", None)
                mso_values_remove.append("password")
            elif auth_password == "" and "password" in mso_values and "password" not in proposed_payload:
                mso_values.pop("password", None)

            if (peer_prefix_uuid == "" or peer_prefix_is_empty) and "peerPrefixRef" in proposed_payload:
                mso_values_remove.append("peerPrefixRef")
                mso_values.pop("peerPrefixRef", None)
                proposed_payload["peerPrefixRef"] = ""
            elif (peer_prefix_uuid == "" or peer_prefix_is_empty) and "peerPrefixRef" not in proposed_payload:
                mso_values.pop("peerPrefixRef", None)

            if (import_route_map_uuid == "" or import_route_map_is_empty) and "importRouteMapRef" in proposed_payload:
                mso_values_remove.append("importRouteMapRef")
                mso_values.pop("importRouteMapRef", None)
                proposed_payload["importRouteMapRef"] = ""
            elif (import_route_map_uuid == "" or import_route_map_is_empty) and "importRouteMapRef" not in proposed_payload:
                mso_values.pop("importRouteMapRef", None)

            if (export_route_map_uuid == "" or export_route_map_is_empty) and "exportRouteMapRef" in proposed_payload:
                mso_values_remove.append("exportRouteMapRef")
                mso_values.pop("exportRouteMapRef", None)
                proposed_payload["exportRouteMapRef"] = ""
            elif (export_route_map_uuid == "" or export_route_map_is_empty) and "exportRouteMapRef" not in proposed_payload:
                mso_values.pop("exportRouteMapRef", None)

            # BGP Controls
            if bgp_controls is not None:
                if bgp_controls.get("state") == "disabled" and proposed_payload.get("bgpControls"):
                    mso_values_remove.append("bgpControls")
                    mso_values.pop("bgpControls", None)
                    proposed_payload["bgpControls"] = {}
                elif bgp_controls.get("state") != "disabled":
                    if not proposed_payload.get("bgpControls"):
                        mso_values["bgpControls"] = dict()
                    mso_values[("bgpControls", "allowSelfAS")] = bgp_controls.get("allow_self_as")
                    mso_values[("bgpControls", "asOverride")] = bgp_controls.get("override_as")
                    mso_values[("bgpControls", "disablePeerASCheck")] = bgp_controls.get("disabled_peer_as_check")
                    mso_values[("bgpControls", "nextHopSelf")] = bgp_controls.get("next_hop_self")
                    mso_values[("bgpControls", "sendCommunity")] = bgp_controls.get("send_community")
                    mso_values[("bgpControls", "sendExtendedCommunity")] = bgp_controls.get("send_extended_community")
                    mso_values[("bgpControls", "sendDomainPath")] = bgp_controls.get("send_domain_path")

            # Peer Controls
            if peer_controls is not None:
                if peer_controls.get("state") == "disabled" and proposed_payload.get("peerControls"):
                    mso_values_remove.append("peerControls")
                    mso_values.pop("peerControls", None)
                    proposed_payload["peerControls"] = {}
                elif peer_controls.get("state") != "disabled":
                    if not proposed_payload.get("peerControls"):
                        mso_values["peerControls"] = dict()

                    mso_values[("peerControls", "bfd")] = peer_controls.get("bfd")
                    mso_values[("peerControls", "disableConnectedCheck")] = peer_controls.get("disable_peer_connected_check")

            # Address Type Controls
            if address_families is not None:
                if address_families.get("state") == "disabled" and proposed_payload.get("addressTypeControls"):
                    mso_values_remove.append("addressTypeControls")
                    mso_values.pop("addressTypeControls", None)
                    proposed_payload["addressTypeControls"] = {}
                elif address_families.get("state") != "disabled":
                    if not proposed_payload.get("addressTypeControls"):
                        mso_values["addressTypeControls"] = dict()

                    mso_values[("addressTypeControls", "afMast")] = address_families.get("multicast")
                    mso_values[("addressTypeControls", "afUcast")] = address_families.get("unicast")

            # Private AS Controls
            if private_as_controls is not None:
                if private_as_controls.get("state") == "disabled" and proposed_payload.get("privateASControls"):
                    mso_values_remove.append("privateASControls")
                    mso_values.pop("privateASControls", None)
                    proposed_payload["privateASControls"] = {}
                elif private_as_controls.get("state") != "disabled":
                    if not proposed_payload.get("privateASControls"):
                        mso_values["privateASControls"] = dict()

                    mso_values[("privateASControls", "removeAll")] = private_as_controls.get("remove_all")
                    mso_values[("privateASControls", "replaceWithLocalAS")] = private_as_controls.get("replace_with_local_as")

                    if private_as_controls.get("remove_all") or private_as_controls.get("replace_with_local_as"):
                        mso_values[("privateASControls", "removeExclusive")] = True

            mso.sanitize(proposed_payload, collate=False, required=required_keys)
            append_update_ops_data(ops, proposed_payload, bgp_peer_path, mso_values, mso_values_remove)
    elif state == "absent":
        if mso.existing:
            ops.append(dict(op="remove", path=bgp_peer_path))

    if not module.check_mode and ops:
        mso_template.template = mso.request(mso_template.template_path, method="PATCH", data=ops)
        l3out_object = mso_template.get_l3out_object(uuid=l3out_uuid, name=l3out, fail_module=True)
        parent_object = mso_template.get_parent_details_for_nested_object_in_l3out(mso_templates, l3out_object)
        match = get_bgp_peer_by_address(mso_template, parent_object[0].details.get("bgpPeers", []), ipv4_addr=ipv4_addr, ipv6_addr=ipv6_addr)
        if match:
            if match.details.get("password", {}).get("ref"):
                match.details["password"].pop("ref", None)
            set_bgp_peer_details(mso_template, parent_type, parent_object[0].details, match.details)
            mso.existing = match.details  # When the state is present
        else:
            mso.existing = {}  # When the state is absent
    elif module.check_mode and state != "query":  # When the state is present/absent with check mode
        set_bgp_peer_details(mso_template, parent_type, parent_object.details, mso.proposed)
        mso.existing = mso.proposed if state == "present" else {}

    mso.exit_json()


def get_bgp_peer_by_address(mso_template, bgp_peers, ipv4_addr=None, ipv6_addr=None, fail_module=False):
    if bgp_peers and (ipv4_addr or ipv6_addr):  # Query a specific object
        kv_list = []
        if ipv4_addr:
            kv_list.append(KVPair("peerAddressV4", ipv4_addr))
        if ipv6_addr:
            kv_list.append(KVPair("peerAddressV6", ipv6_addr))
        return mso_template.get_object_by_key_value_pairs("L3Out BGP Peer", bgp_peers, kv_list, fail_module)
    return bgp_peers  # Query all objects


def set_bgp_peer_details(mso_template, parent_type, parent_object, match):
    reference_dict = {
        "peerPrefix": {
            "name": "peerPrefixName",
            "reference": "peerPrefixRef",
            "type": "bgpPeerPrefixPol",
            "template": "peerPrefixTemplateName",
            "templateId": "peerPrefixTemplateId",
        },
        "importRouteMap": {
            "name": "importRouteMapName",
            "reference": "importRouteMapRef",
            "type": "routeMap",
            "template": "importRouteMapTemplateName",
            "templateId": "importRouteMapTemplateId",
        },
        "exportRouteMap": {
            "name": "exportRouteMapName",
            "reference": "exportRouteMapRef",
            "type": "routeMap",
            "template": "exportRouteMapTemplateName",
            "templateId": "exportRouteMapTemplateId",
        },
    }

    mso_template.update_config_with_template_and_references(match, reference_dict)
    mso_template.set_parent_details_for_nested_object_in_l3out(parent_type, parent_object, match)
    return match


if __name__ == "__main__":
    main()



#!/usr/bin/python
#
# Copyright (c) 2018 Yuwei Zhou, <yuwzho@microsoft.com>
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: azure_rm_servicebustopic
version_added: "0.1.2"
short_description: Manage Azure Service Bus
description:
    - Create, update or delete an Azure Service Bus topics.
options:
    resource_group:
        description:
            - Name of resource group.
        required: true
        type: str
    name:
        description:
            - Name of the topic.
        required: true
        type: str
    namespace:
        description:
            - Servicebus namespace name.
            - A namespace is a scoping container for all messaging components.
            - Multipletopics can reside within a single namespace.
        type: str
        required: true
    state:
        description:
            - Assert the state of the topic. Use C(present) to create or update and use C(absent) to delete.
        default: present
        type: str
        choices:
            - absent
            - present
    auto_delete_on_idle_in_seconds:
        description:
            - Time idle interval after which a topic is automatically deleted.
            - The minimum duration is 5 minutes.
        type: int
    default_message_time_to_live_seconds:
        description:
            - Default message timespan to live value.
            - This is the duration after which the message expires, starting from when the message is sent to Service Bus.
            - This is the default value used when TimeToLive is not set on a message itself.
        type: int
    enable_batched_operations:
        description:
            - Value that indicates whether server-side batched operations are enabled.
        type: bool
    enable_express:
        description:
            - Value that indicates whether Express Entities are enabled.
            - An express topic holds a message in memory temporarily before writing it to persistent storage.
        type: bool
    enable_partitioning:
        description:
            - A value that indicates whether the topic is to be partitioned across multiple message brokers.
        type: bool
    max_message_size_in_kb:
        description:
            - Maximum size (in KB) of the message payload that can be accepted by the queue.
            - This property is only used in Premium today and default is 1024.
        type: int
    max_size_in_mb:
        description:
            - The maximum size of the topic in megabytes, which is the size of memory allocated for the topic.
        type: int
    requires_duplicate_detection:
        description:
            -  A value indicating if this topic requires duplicate detection.
        type: bool
    duplicate_detection_time_in_seconds:
        description:
            - TimeSpan structure that defines the duration of the duplicate detection history.
        type: int
    support_ordering:
        description:
            - Value that indicates whether the topic supports ordering.
        type: bool
    status:
        description:
            - Status of the entity.
        type: str
        choices:
            - active
            - disabled
            - send_disabled
            - receive_disabled

extends_documentation_fragment:
    - azure.azcollection.azure

author:
    - Yuwei Zhou (@yuwzho)

'''

EXAMPLES = '''
- name: Create a topic
  azure_rm_servicebustopic:
      name: subtopic
      resource_group: myResourceGroup
      namespace: bar
      duplicate_detection_time_in_seconds: 600
'''
RETURN = '''
id:
    description:
        - Current state of the topic.
    returned: success
    type: str
    sample: "/subscriptions/xxx...xxx/resourceGroups/myResourceGroup/providers/Microsoft.ServiceBus/namespaces/nsb57dc95979/topics/topicb57dc95979"
'''

try:
    from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase
except ImportError:
    # This is handled in azure_rm_common
    pass

from ansible.module_utils.common.dict_transformations import _snake_to_camel, _camel_to_snake
from ansible.module_utils._text import to_native
from datetime import datetime, timedelta


duration_spec_map = dict(
    default_message_time_to_live='default_message_time_to_live_seconds',
    duplicate_detection_history_time_window='duplicate_detection_time_in_seconds',
    auto_delete_on_idle='auto_delete_on_idle_in_seconds'
)


sas_policy_spec = dict(
    state=dict(type='str', default='present', choices=['present', 'absent']),
    name=dict(type='str', required=True),
    regenerate_key=dict(type='bool'),
    rights=dict(type='str', choices=['manage', 'listen', 'send', 'listen_send'])
)


class AzureRMServiceBusTopic(AzureRMModuleBase):

    def __init__(self):

        self.module_arg_spec = dict(
            auto_delete_on_idle_in_seconds=dict(type='int'),
            default_message_time_to_live_seconds=dict(type='int'),
            duplicate_detection_time_in_seconds=dict(type='int'),
            enable_batched_operations=dict(type='bool'),
            enable_express=dict(type='bool'),
            enable_partitioning=dict(type='bool'),
            max_size_in_mb=dict(type='int'),
            max_message_size_in_kb=dict(type='int'),
            name=dict(type='str', required=True),
            namespace=dict(type='str', required=True),
            requires_duplicate_detection=dict(type='bool'),
            resource_group=dict(type='str', required=True),
            state=dict(type='str', default='present', choices=['present', 'absent']),
            status=dict(type='str',
                        choices=['active', 'disabled', 'send_disabled', 'receive_disabled']),
            support_ordering=dict(type='bool')
        )

        self.resource_group = None
        self.name = None
        self.state = None
        self.namespace = None
        self.auto_delete_on_idle_in_seconds = None
        self.default_message_time_to_live_seconds = None
        self.enable_batched_operations = None
        self.enable_express = None
        self.enable_partitioning = None
        self.max_size_in_mb = None
        self.requires_duplicate_detection = None
        self.status = None
        self.support_ordering = None
        self.max_message_size_in_kb = None

        self.results = dict(
            changed=False,
            id=None
        )

        super(AzureRMServiceBusTopic, self).__init__(self.module_arg_spec,
                                                     supports_tags=False,
                                                     supports_check_mode=True)

    def exec_module(self, **kwargs):

        for key in list(self.module_arg_spec.keys()):
            setattr(self, key, kwargs[key])

        changed = False
        original = self.get()
        if self.state == 'present':
            # Create the resource instance
            params = dict(
                enable_batched_operations=self.enable_batched_operations,
                enable_express=self.enable_express,
                enable_partitioning=self.enable_partitioning,
                max_size_in_megabytes=self.max_size_in_mb,
                max_message_size_in_kilobytes=self.max_message_size_in_kb,
                support_ordering=self.support_ordering
            )
            if self.status:
                params['status'] = self.servicebus_models.EntityStatus(str.capitalize(_snake_to_camel(self.status)))
            for k, v in duration_spec_map.items():
                seconds = getattr(self, v)
                if seconds:
                    params[k] = timedelta(seconds=seconds)

            instance = self.servicebus_models.SBTopic(**params)
            result = original
            if not original:
                changed = True
                result = instance
            else:
                result = original
                attribute_map = set(self.servicebus_models.SBTopic._attribute_map.keys()) - set(self.servicebus_models.SBTopic._validation.keys())
                for attribute in attribute_map:
                    value = getattr(instance, attribute)
                    if value and value != getattr(original, attribute):
                        changed = True
            if changed and not self.check_mode:
                result = self.create_or_update(instance)
            self.results = self.to_dict(result)
        elif original:
            changed = True
            if not self.check_mode:
                self.delete()
                self.results['deleted'] = True

        self.results['changed'] = changed
        return self.results

    def create_or_update(self, param):
        try:
            client = self._get_client()
            return client.create_or_update(self.resource_group, self.namespace, self.name, param)
        except Exception as exc:
            self.fail('Error creating or updating topic {0} - {1}'.format(self.name, str(exc.inner_exception) or str(exc)))

    def delete(self):
        try:
            client = self._get_client()
            client.delete(self.resource_group, self.namespace, self.name)
            return True
        except Exception as exc:
            self.fail("Error deleting topic {0} - {1}".format(self.name, str(exc)))

    def _get_client(self):
        return self.servicebus_client.topics

    def get(self):
        try:
            client = self._get_client()
            return client.get(self.resource_group, self.namespace, self.name)
        except Exception:
            return None

    def to_dict(self, instance):
        result = dict()
        attribute_map = self.servicebus_models.SBTopic._attribute_map
        for attribute in attribute_map.keys():
            value = getattr(instance, attribute)
            if not value:
                continue
            if attribute_map[attribute]['type'] == 'duration':
                if is_valid_timedelta(value):
                    key = duration_spec_map.get(attribute) or attribute
                    result[key] = int(value.total_seconds())
            elif attribute == 'status':
                result['status'] = _camel_to_snake(value)
            elif isinstance(value, self.servicebus_models.MessageCountDetails):
                result[attribute] = value.as_dict()
            elif isinstance(value, self.servicebus_models.SBSku):
                result[attribute] = value.name.lower()
            elif isinstance(value, datetime):
                result[attribute] = str(value)
            elif isinstance(value, str):
                result[attribute] = to_native(value)
            elif attribute == 'max_size_in_megabytes':
                result['max_size_in_mb'] = value
            elif attribute == 'max_message_size_in_kilobyte':
                result['max_message_size_in_kb'] = value
            else:
                result[attribute] = value
        return result


def is_valid_timedelta(value):
    if value == timedelta(10675199, 10085, 477581):
        return None
    return value


def main():
    AzureRMServiceBusTopic()


if __name__ == '__main__':
    main()



#!/usr/bin/python

# Copyright (c) 2016, Jiri Tyr <jiri.tyr@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: jenkins_plugin
author: Jiri Tyr (@jtyr)
short_description: Add or remove Jenkins plugin
description:
  - Ansible module which helps to manage Jenkins plugins.
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none

options:
  group:
    type: str
    description:
      - GID or name of the Jenkins group on the OS.
    default: jenkins
  jenkins_home:
    type: path
    description:
      - Home directory of the Jenkins user.
    default: /var/lib/jenkins
  mode:
    type: raw
    description:
      - File mode applied on versioned plugins.
    default: '0644'
  name:
    type: str
    description:
      - Plugin name.
    required: true
  owner:
    type: str
    description:
      - UID or name of the Jenkins user on the OS.
    default: jenkins
  state:
    type: str
    description:
      - Desired plugin state.
      - If set to V(latest), the check for new version is performed every time. This is suitable to keep the plugin up-to-date.
    choices: [absent, present, pinned, unpinned, enabled, disabled, latest]
    default: present
  timeout:
    type: int
    description:
      - Server connection timeout in secs.
    default: 30
  updates_expiration:
    type: int
    description:
      - Number of seconds after which a new copy of the C(update-center.json) file is downloaded. This is used to avoid the
        need to download the plugin to calculate its checksum when O(state=latest) is specified.
      - Set it to V(0) if no cache file should be used. In that case, the plugin file is always downloaded to calculate its
        checksum when O(state=latest) is specified.
    default: 86400
  updates_url:
    type: list
    elements: str
    description:
      - A list of base URL(s) to retrieve C(update-center.json), and direct plugin files from.
      - This can be a list since community.general 3.3.0.
    default: ['https://updates.jenkins.io', 'http://mirrors.jenkins.io']
  updates_url_username:
    description:
      - If using a custom O(updates_url), set this as the username of the user with access to the URL.
      - If the custom O(updates_url) does not require authentication, this can be left empty.
    type: str
    version_added: 11.2.0
  updates_url_password:
    description:
      - If using a custom O(updates_url), set this as the password of the user with access to the URL.
      - If the custom O(updates_url) does not require authentication, this can be left empty.
    type: str
    version_added: 11.2.0
  update_json_url_segment:
    type: list
    elements: str
    description:
      - A list of URL segment(s) to retrieve the update center JSON file from.
    default: ['update-center.json', 'updates/update-center.json']
    version_added: 3.3.0
  plugin_versions_url_segment:
    type: list
    elements: str
    description:
      - A list of URL segment(s) to retrieve the plugin versions JSON file from.
    default: ['plugin-versions.json', 'current/plugin-versions.json']
    version_added: 11.2.0
  latest_plugins_url_segments:
    type: list
    elements: str
    description:
      - Path inside the O(updates_url) to get latest plugins from.
    default: ['latest']
    version_added: 3.3.0
  versioned_plugins_url_segments:
    type: list
    elements: str
    description:
      - Path inside the O(updates_url) to get specific version of plugins from.
    default: ['download/plugins', 'plugins']
    version_added: 3.3.0
  url:
    type: str
    description:
      - URL of the Jenkins server.
    default: http://localhost:8080
  version:
    type: str
    description:
      - Plugin version number.
      - If this option is specified, all plugin dependencies must be installed manually.
      - It might take longer to verify that the correct version is installed. This is especially true if a specific version
        number is specified.
      - Quote the version to prevent the value to be interpreted as float. For example if V(1.20) would be unquoted, it would
        become V(1.2).
  with_dependencies:
    description:
      - Defines whether to install plugin dependencies.
      - In earlier versions, this option had no effect when a specific O(version) was set.
        Since community.general 11.2.0, dependencies are also installed for versioned plugins.
    type: bool
    default: true

notes:
  - Plugin installation should be run under root or the same user which owns the plugin files on the disk. Only if the plugin
    is not installed yet and no version is specified, the API installation is performed which requires only the Web UI credentials.
  - It is necessary to notify the handler or call the M(ansible.builtin.service) module to restart the Jenkins service after
    a new plugin was installed.
  - Pinning works only if the plugin is installed and Jenkins service was successfully restarted after the plugin installation.
  - It is not possible to run the module remotely by changing the O(url) parameter to point to the Jenkins server. The module
    must be used on the host where Jenkins runs as it needs direct access to the plugin files.
  - If using a custom O(updates_url), ensure that the URL provides a C(plugin-versions.json) file.
    This file must include metadata for all available plugin versions to support version compatibility resolution.
    The file should be in the same format as the one provided by Jenkins update center (https://updates.jenkins.io/current/plugin-versions.json).
extends_documentation_fragment:
  - ansible.builtin.url
  - ansible.builtin.files
  - community.general._attributes
"""

EXAMPLES = r"""
- name: Install plugin
  community.general.jenkins_plugin:
    name: build-pipeline-plugin

- name: Install plugin without its dependencies
  community.general.jenkins_plugin:
    name: build-pipeline-plugin
    with_dependencies: false

- name: Make sure the plugin is always up-to-date
  community.general.jenkins_plugin:
    name: token-macro
    state: latest

- name: Install specific version of the plugin
  community.general.jenkins_plugin:
    name: token-macro
    version: "1.15"

- name: Pin the plugin
  community.general.jenkins_plugin:
    name: token-macro
    state: pinned

- name: Unpin the plugin
  community.general.jenkins_plugin:
    name: token-macro
    state: unpinned

- name: Enable the plugin
  community.general.jenkins_plugin:
    name: token-macro
    state: enabled

- name: Disable the plugin
  community.general.jenkins_plugin:
    name: token-macro
    state: disabled

- name: Uninstall plugin
  community.general.jenkins_plugin:
    name: build-pipeline-plugin
    state: absent

#
# Example of how to authenticate
#
- name: Install plugin
  community.general.jenkins_plugin:
    name: build-pipeline-plugin
    url_username: admin
    url_password: p4ssw0rd
    url: http://localhost:8888

#
# Example of how to authenticate with serverless deployment
#
- name: Update plugins on ECS Fargate Jenkins instance
  community.general.jenkins_plugin:
    # plugin name and version
    name: ws-cleanup
    version: '0.45'
    # Jenkins home path mounted on ec2-helper VM (example)
    jenkins_home: "/mnt/{{ jenkins_instance }}"
    # matching the UID/GID to one in official Jenkins image
    owner: 1000
    group: 1000
    # Jenkins instance URL and admin credentials
    url: "https://{{ jenkins_instance }}.com/"
    url_username: admin
    url_password: p4ssw0rd
  # make module work from EC2 which has local access
  # to EFS mount as well as Jenkins URL
  delegate_to: ec2-helper
  vars:
    jenkins_instance: foobar

#
# Example of a Play which handles Jenkins restarts during the state changes
#
- name: Jenkins Master play
  hosts: jenkins-master
  vars:
    my_jenkins_plugins:
      token-macro:
        enabled: true
      build-pipeline-plugin:
        version: "1.4.9"
        pinned: false
        enabled: true
  tasks:
    - name: Install plugins without a specific version
      community.general.jenkins_plugin:
        name: "{{ item.key }}"
      register: my_jenkins_plugin_unversioned
      when: >
        'version' not in item.value
      with_dict: "{{ my_jenkins_plugins }}"

    - name: Install plugins with a specific version
      community.general.jenkins_plugin:
        name: "{{ item.key }}"
        version: "{{ item.value['version'] }}"
      register: my_jenkins_plugin_versioned
      when: >
        'version' in item.value
      with_dict: "{{ my_jenkins_plugins }}"

    - name: Initiate the fact
      ansible.builtin.set_fact:
        jenkins_restart_required: false

    - name: Check if restart is required by any of the versioned plugins
      ansible.builtin.set_fact:
        jenkins_restart_required: true
      when: item.changed
      with_items: "{{ my_jenkins_plugin_versioned.results }}"

    - name: Check if restart is required by any of the unversioned plugins
      ansible.builtin.set_fact:
        jenkins_restart_required: true
      when: item.changed
      with_items: "{{ my_jenkins_plugin_unversioned.results }}"

    - name: Restart Jenkins if required
      ansible.builtin.service:
        name: jenkins
        state: restarted
      when: jenkins_restart_required

    - name: Wait for Jenkins to start up
      ansible.builtin.uri:
        url: http://localhost:8080
        status_code: 200
        timeout: 5
      register: jenkins_service_status
      # Keep trying for 5 mins in 5 sec intervals
      retries: 60
      delay: 5
      until: >
        'status' in jenkins_service_status and
        jenkins_service_status['status'] == 200
      when: jenkins_restart_required

    - name: Reset the fact
      ansible.builtin.set_fact:
        jenkins_restart_required: false
      when: jenkins_restart_required

    - name: Plugin pinning
      community.general.jenkins_plugin:
        name: "{{ item.key }}"
        state: "{{ 'pinned' if item.value['pinned'] else 'unpinned'}}"
      when: >
        'pinned' in item.value
      with_dict: "{{ my_jenkins_plugins }}"

    - name: Plugin enabling
      community.general.jenkins_plugin:
        name: "{{ item.key }}"
        state: "{{ 'enabled' if item.value['enabled'] else 'disabled'}}"
      when: >
        'enabled' in item.value
      with_dict: "{{ my_jenkins_plugins }}"
"""

RETURN = r"""
plugin:
  description: Plugin name.
  returned: success
  type: str
  sample: build-pipeline-plugin
state:
  description: State of the target, after execution.
  returned: success
  type: str
  sample: "present"
"""

import hashlib
import json
import os
import tempfile
import time
from collections import OrderedDict
from http import cookiejar
from urllib.parse import urlencode

from ansible.module_utils.basic import AnsibleModule, to_bytes
from ansible.module_utils.urls import basic_auth_header, fetch_url, url_argument_spec

from ansible_collections.community.general.plugins.module_utils._jenkins import download_updates_file


class FailedInstallingWithPluginManager(Exception):
    pass


class JenkinsPlugin:
    def __init__(self, module):
        # To be able to call fail_json
        self.module = module

        # Shortcuts for the params
        self.params = self.module.params
        self.url = self.params["url"]
        self.timeout = self.params["timeout"]

        # Authentication for non-Jenkins calls
        self.updates_url_credentials = {}
        if self.params.get("updates_url_username") and self.params.get("updates_url_password"):
            self.updates_url_credentials["Authorization"] = basic_auth_header(
                self.params["updates_url_username"], self.params["updates_url_password"]
            )

        # Crumb
        self.crumb = {}

        # Authentication for Jenkins calls
        if self.params.get("url_username") and self.params.get("url_password"):
            self.crumb["Authorization"] = basic_auth_header(self.params["url_username"], self.params["url_password"])

        # Cookie jar for crumb session
        self.cookies = None

        if self._csrf_enabled():
            self.cookies = cookiejar.LWPCookieJar()
            self._get_crumb()

        # Get list of installed plugins
        self._get_installed_plugins()

    def _csrf_enabled(self):
        csrf_data = self._get_json_data(f"{self.url}/api/json", "CSRF")

        if "useCrumbs" not in csrf_data:
            self.module.fail_json(msg="Required fields not found in the Crumbs response.", details=csrf_data)

        return csrf_data["useCrumbs"]

    def _get_json_data(self, url, what, **kwargs):
        # Get the JSON data
        r = self._get_url_data(url, what, **kwargs)

        # Parse the JSON data
        try:
            json_data = json.loads(r.read())
        except Exception as e:
            self.module.fail_json(msg=f"Cannot parse {what} JSON data.", details=f"{e}")

        return json_data

    def _get_urls_data(self, urls, what=None, msg_status=None, msg_exception=None, **kwargs):
        # Compose default messages
        if msg_status is None:
            msg_status = f"Cannot get {what}"

        if msg_exception is None:
            msg_exception = f"Retrieval of {what} failed."

        errors = {}
        for url in urls:
            err_msg = None
            try:
                self.module.debug(f"fetching url: {url}")

                is_jenkins_call = url.startswith(self.url)
                self.module.params["force_basic_auth"] = is_jenkins_call

                response, info = fetch_url(
                    self.module,
                    url,
                    timeout=self.timeout,
                    cookies=self.cookies,
                    headers=self.crumb if is_jenkins_call else self.updates_url_credentials or self.crumb,
                    **kwargs,
                )
                if info["status"] == 200:
                    return response
                else:
                    err_msg = f"{msg_status}. fetching url {url} failed. response code: {info['status']}"
                    if info["status"] > 400:  # extend error message
                        err_msg = f"{err_msg}. response body: {info['body']}"
            except Exception as e:
                err_msg = f"{msg_status}. fetching url {url} failed. error msg: {e}"
            finally:
                if err_msg is not None:
                    self.module.debug(err_msg)
                    errors[url] = err_msg

        # failed on all urls
        self.module.fail_json(msg=msg_exception, details=errors)

    def _get_url_data(self, url, what=None, msg_status=None, msg_exception=None, dont_fail=False, **kwargs):
        # Compose default messages
        if msg_status is None:
            msg_status = f"Cannot get {what}"

        if msg_exception is None:
            msg_exception = f"Retrieval of {what} failed."

        # Get the URL data
        try:
            is_jenkins_call = url.startswith(self.url)
            self.module.params["force_basic_auth"] = is_jenkins_call

            response, info = fetch_url(
                self.module,
                url,
                timeout=self.timeout,
                cookies=self.cookies,
                headers=self.crumb if is_jenkins_call else self.updates_url_credentials or self.crumb,
                **kwargs,
            )

            if info["status"] != 200:
                if dont_fail:
                    raise FailedInstallingWithPluginManager(info["msg"])
                else:
                    self.module.fail_json(msg=msg_status, details=info["msg"])
        except Exception as e:
            if dont_fail:
                raise FailedInstallingWithPluginManager(e) from e
            else:
                self.module.fail_json(msg=msg_exception, details=f"{e}")

        return response

    def _get_crumb(self):
        crumb_data = self._get_json_data(f"{self.url}/crumbIssuer/api/json", "Crumb")

        if "crumbRequestField" in crumb_data and "crumb" in crumb_data:
            self.crumb[crumb_data["crumbRequestField"]] = crumb_data["crumb"]
        else:
            self.module.fail_json(msg="Required fields not found in the Crum response.", details=crumb_data)

    def _get_installed_plugins(self):
        plugins_data = self._get_json_data(f"{self.url}/pluginManager/api/json?depth=1", "list of plugins")

        # Check if we got valid data
        if "plugins" not in plugins_data:
            self.module.fail_json(msg="No valid plugin data found.")

        # Create final list of installed/pined plugins
        self.is_installed = False
        self.is_pinned = False
        self.is_enabled = False
        self.installed_plugins = plugins_data["plugins"]

        for p in plugins_data["plugins"]:
            if p["shortName"] == self.params["name"]:
                self.is_installed = True

                if p["pinned"]:
                    self.is_pinned = True

                if p["enabled"]:
                    self.is_enabled = True

                break

    def _install_dependencies(self):
        dependencies = self._get_versioned_dependencies()
        self.dependencies_states = []

        for dep_name, dep_version in dependencies.items():
            if not any(p["shortName"] == dep_name and p["version"] == dep_version for p in self.installed_plugins):
                dep_params = self.params.copy()
                dep_params["name"] = dep_name
                dep_params["version"] = dep_version
                dep_module = AnsibleModule(
                    argument_spec=self.module.argument_spec, supports_check_mode=self.module.check_mode
                )
                dep_module.params = dep_params
                dep_plugin = JenkinsPlugin(dep_module)
                if not dep_plugin.install():
                    self.dependencies_states.append({"name": dep_name, "version": dep_version, "state": "absent"})
                else:
                    self.dependencies_states.append({"name": dep_name, "version": dep_version, "state": "present"})
            else:
                self.dependencies_states.append({"name": dep_name, "version": dep_version, "state": "present"})

    def _install_with_plugin_manager(self):
        if not self.module.check_mode:
            # Install the plugin (with dependencies)
            install_script = (
                f"""d = Jenkins.instance.updateCenter.getPlugin("{self.params["name"]}").deploy(); d.get();"""
            )

            if self.params["with_dependencies"]:
                install_script = (
                    f'Jenkins.instance.updateCenter.getPlugin("{self.params["name"]}")'
                    f".getNeededDependencies().each{{it.deploy()}}; {install_script}"
                )

            script_data = {"script": install_script}
            data = urlencode(script_data)

            # Send the installation request
            self._get_url_data(
                f"{self.url}/scriptText",
                msg_status="Cannot install plugin.",
                msg_exception="Plugin installation has failed.",
                data=data,
                dont_fail=True,
            )

            hpi_file = f"{self.params['jenkins_home']}/plugins/{self.params['name']}.hpi"

            if os.path.isfile(hpi_file):
                os.remove(hpi_file)

    def install(self):
        changed = False
        plugin_file = f"{self.params['jenkins_home']}/plugins/{self.params['name']}.jpi"

        if not self.is_installed and self.params["version"] in [None, "latest"]:
            try:
                self._install_with_plugin_manager()
                changed = True
            except FailedInstallingWithPluginManager:  # Fallback to manually downloading the plugin
                pass

        if not changed:
            # Check if the plugin directory exists
            if not os.path.isdir(self.params["jenkins_home"]):
                self.module.fail_json(msg="Jenkins home directory doesn't exist.")

            checksum_old = None
            if os.path.isfile(plugin_file):
                # Make the checksum of the currently installed plugin
                with open(plugin_file, "rb") as plugin_fh:
                    plugin_content = plugin_fh.read()
                checksum_old = hashlib.sha1(plugin_content).hexdigest()

            # Install dependencies
            if self.params["with_dependencies"]:
                self._install_dependencies()

            if self.params["version"] in [None, "latest"]:
                # Take latest version
                plugin_urls = self._get_latest_plugin_urls()
            else:
                # Take specific version
                plugin_urls = self._get_versioned_plugin_urls()
            if (
                self.params["updates_expiration"] == 0
                or self.params["version"] not in [None, "latest"]
                or checksum_old is None
            ):
                # Download the plugin file directly
                r = self._download_plugin(plugin_urls)

                # Write downloaded plugin into file if checksums don't match
                if checksum_old is None:
                    # No previously installed plugin
                    if not self.module.check_mode:
                        self._write_file(plugin_file, r)

                    changed = True
                else:
                    # Get data for the MD5
                    data = r.read()

                    # Make new checksum
                    checksum_new = hashlib.sha1(data).hexdigest()

                    # If the checksum is different from the currently installed
                    # plugin, store the new plugin
                    if checksum_old != checksum_new:
                        if not self.module.check_mode:
                            self._write_file(plugin_file, data)

                        changed = True
            elif self.params["version"] == "latest":
                # Check for update from the updates JSON file
                plugin_data = self._download_updates()

                # If the latest version changed, download it
                if checksum_old != to_bytes(plugin_data["sha1"]):
                    if not self.module.check_mode:
                        r = self._download_plugin(plugin_urls)
                        self._write_file(plugin_file, r)

                    changed = True

        # Change file attributes if needed
        if os.path.isfile(plugin_file):
            params = {"dest": plugin_file}
            params.update(self.params)
            file_args = self.module.load_file_common_arguments(params)

            if not self.module.check_mode:
                # Not sure how to run this in the check mode
                changed = self.module.set_fs_attributes_if_different(file_args, changed)
            else:
                # See the comment above
                changed = True

        return changed

    def _get_latest_plugin_urls(self):
        urls = []
        for base_url in self.params["updates_url"]:
            for update_segment in self.params["latest_plugins_url_segments"]:
                urls.append(f"{base_url}/{update_segment}/{self.params['name']}.hpi")
        return urls

    def _get_latest_compatible_plugin_version(self, plugin_name=None):
        if not hasattr(self, "jenkins_version"):
            self.module.params["force_basic_auth"] = True
            resp, info = fetch_url(self.module, self.url)
            raw_version = info.get("x-jenkins")
            self.jenkins_version = self.parse_version(raw_version)
        name = plugin_name or self.params["name"]
        cache_path = f"{self.params['jenkins_home']}/ansible_jenkins_plugin_cache.json"
        plugin_version_urls = []
        for base_url in self.params["updates_url"]:
            for update_json in self.params["plugin_versions_url_segment"]:
                plugin_version_urls.append(f"{base_url}/{update_json}")

        try:  # Check if file is saved localy
            if os.path.exists(cache_path):
                file_mtime = os.path.getmtime(cache_path)
            else:
                file_mtime = 0

            now = time.time()
            if now - file_mtime >= 86400:
                response = self._get_urls_data(plugin_version_urls, what="plugin-versions.json")
                plugin_data = json.loads(response.read(), object_pairs_hook=OrderedDict)

                # Save it to file for next time
                with open(cache_path, "w") as f:
                    json.dump(plugin_data, f)

            with open(cache_path) as f:
                plugin_data = json.load(f)

        except Exception as e:
            if os.path.exists(cache_path):
                os.remove(cache_path)
            self.module.fail_json(msg="Failed to parse plugin-versions.json", details=f"{e}")

        plugin_versions = plugin_data.get("plugins", {}).get(name)
        if not plugin_versions:
            self.module.fail_json(msg=f"Plugin '{name}' not found.")

        sorted_versions = list(reversed(plugin_versions.items()))

        for idx, (version_title, version_info) in enumerate(sorted_versions):
            required_core = version_info.get("requiredCore", "0.0")
            if self.parse_version(required_core) <= self.jenkins_version:
                return "latest" if idx == 0 else version_title

        self.module.warn(f"No compatible version found for plugin '{name}'. Installing latest version.")
        return "latest"

    def _get_versioned_plugin_urls(self):
        urls = []
        for base_url in self.params["updates_url"]:
            for versioned_segment in self.params["versioned_plugins_url_segments"]:
                urls.append(
                    f"{base_url}/{versioned_segment}/{self.params['name']}/{self.params['version']}/{self.params['name']}.hpi"
                )
        return urls

    def _get_update_center_urls(self):
        urls = []
        for base_url in self.params["updates_url"]:
            for update_json in self.params["update_json_url_segment"]:
                urls.append(f"{base_url}/{update_json}")
        return urls

    def _get_versioned_dependencies(self):
        # Get dependencies for the specified plugin version
        plugin_data = self._download_updates()["dependencies"]

        dependencies_info = {
            dep["name"]: self._get_latest_compatible_plugin_version(dep["name"])
            for dep in plugin_data
            if not dep.get("optional", False)
        }

        return dependencies_info

    def _download_updates(self):
        try:
            updates_file, download_updates = download_updates_file(self.params["updates_expiration"])
        except OSError as e:
            self.module.fail_json(msg="Cannot create temporal directory.", details=f"{e}")

        # Download the updates file if needed
        if download_updates:
            urls = self._get_update_center_urls()

            # Get the data
            r = self._get_urls_data(
                urls, msg_status="Remote updates not found.", msg_exception="Updates download failed."
            )

            # Write the updates file
            tmp_update_fd, tmp_updates_file = tempfile.mkstemp()
            os.write(tmp_update_fd, r.read())

            try:
                os.close(tmp_update_fd)
            except OSError as e:
                self.module.fail_json(msg=f"Cannot close the tmp updates file {tmp_updates_file}.", details=f"{e}")
        else:
            tmp_updates_file = updates_file

        # Open the updates file
        try:
            f = open(tmp_updates_file, encoding="utf-8")

            # Read only the second line
            dummy = f.readline()
            data = json.loads(f.readline())
        except OSError as e:
            self.module.fail_json(
                msg=f"Cannot open{' temporary' if tmp_updates_file != updates_file else ''} updates file.",
                details=f"{e}",
            )
        except Exception as e:
            self.module.fail_json(
                msg=f"Cannot load JSON data from the{' temporary' if tmp_updates_file != updates_file else ''} updates file.",
                details=f"{e}",
            )

        # Move the updates file to the right place if we could read it
        if tmp_updates_file != updates_file:
            self.module.atomic_move(os.path.abspath(tmp_updates_file), os.path.abspath(updates_file))

        # Check if we have the plugin data available
        if not data.get("plugins", {}).get(self.params["name"]):
            self.module.fail_json(msg="Cannot find plugin data in the updates file.")

        return data["plugins"][self.params["name"]]

    def _download_plugin(self, plugin_urls):
        # Download the plugin

        return self._get_urls_data(plugin_urls, msg_status="Plugin not found.", msg_exception="Plugin download failed.")

    def _write_file(self, f, data):
        # Store the plugin into a temp file and then move it
        tmp_f_fd, tmp_f = tempfile.mkstemp()

        if isinstance(data, (str, bytes)):
            os.write(tmp_f_fd, data)
        else:
            os.write(tmp_f_fd, data.read())

        try:
            os.close(tmp_f_fd)
        except OSError as e:
            self.module.fail_json(msg=f"Cannot close the temporal plugin file {tmp_f}.", details=f"{e}")

        # Move the file onto the right place
        self.module.atomic_move(os.path.abspath(tmp_f), os.path.abspath(f))

    def uninstall(self):
        changed = False

        # Perform the action
        if self.is_installed:
            if not self.module.check_mode:
                self._pm_query("doUninstall", "Uninstallation")

            changed = True

        return changed

    def pin(self):
        return self._pinning("pin")

    def unpin(self):
        return self._pinning("unpin")

    def _pinning(self, action):
        changed = False

        # Check if the plugin is pinned/unpinned
        if action == "pin" and not self.is_pinned or action == "unpin" and self.is_pinned:
            # Perform the action
            if not self.module.check_mode:
                self._pm_query(action, f"{action.capitalize()}ning")

            changed = True

        return changed

    def enable(self):
        return self._enabling("enable")

    def disable(self):
        return self._enabling("disable")

    def _enabling(self, action):
        changed = False

        # Check if the plugin is pinned/unpinned
        if action == "enable" and not self.is_enabled or action == "disable" and self.is_enabled:
            # Perform the action
            if not self.module.check_mode:
                self._pm_query(f"make{action.capitalize()}d", f"{action[:-1].capitalize()}ing")

            changed = True

        return changed

    def _pm_query(self, action, msg):
        url = f"{self.params['url']}/pluginManager/plugin/{self.params['name']}/{action}"

        # Send the request
        self._get_url_data(
            url, msg_status=f"Plugin not found. {url}", msg_exception=f"{msg} has failed.", method="POST"
        )

    @staticmethod
    def parse_version(version_str):
        return tuple(int(x) for x in version_str.split("."))


def main():
    # Module arguments
    argument_spec = url_argument_spec()
    argument_spec.update(
        group=dict(type="str", default="jenkins"),
        jenkins_home=dict(type="path", default="/var/lib/jenkins"),
        mode=dict(default="0644", type="raw"),
        name=dict(type="str", required=True),
        owner=dict(type="str", default="jenkins"),
        state=dict(
            choices=["present", "absent", "pinned", "unpinned", "enabled", "disabled", "latest"], default="present"
        ),
        timeout=dict(default=30, type="int"),
        updates_expiration=dict(default=86400, type="int"),
        updates_url=dict(
            type="list", elements="str", default=["https://updates.jenkins.io", "http://mirrors.jenkins.io"]
        ),
        updates_url_username=dict(type="str"),
        updates_url_password=dict(type="str", no_log=True),
        update_json_url_segment=dict(
            type="list", elements="str", default=["update-center.json", "updates/update-center.json"]
        ),
        plugin_versions_url_segment=dict(
            type="list", elements="str", default=["plugin-versions.json", "current/plugin-versions.json"]
        ),
        latest_plugins_url_segments=dict(type="list", elements="str", default=["latest"]),
        versioned_plugins_url_segments=dict(type="list", elements="str", default=["download/plugins", "plugins"]),
        url=dict(default="http://localhost:8080"),
        url_password=dict(no_log=True),
        version=dict(),
        with_dependencies=dict(default=True, type="bool"),
    )
    # Module settings
    module = AnsibleModule(
        argument_spec=argument_spec,
        add_file_common_args=True,
        supports_check_mode=True,
    )

    # Convert timeout to float
    try:
        module.params["timeout"] = float(module.params["timeout"])
    except ValueError as e:
        module.fail_json(msg=f"Cannot convert {module.params['timeout']} to float.", details=f"{e}")
    # Instantiate the JenkinsPlugin object
    jp = JenkinsPlugin(module)

    # Set version to latest if state is latest
    if module.params["state"] == "latest":
        module.params["state"] = "present"
        module.params["version"] = jp._get_latest_compatible_plugin_version()

    # Set version to latest compatible version if version is latest
    if module.params["version"] == "latest":
        module.params["version"] = jp._get_latest_compatible_plugin_version()

    # Create some shortcuts
    name = module.params["name"]
    state = module.params["state"]

    # Initial change state of the task
    changed = False

    # Perform action depending on the requested state
    if state == "present":
        changed = jp.install()
    elif state == "absent":
        changed = jp.uninstall()
    elif state == "pinned":
        changed = jp.pin()
    elif state == "unpinned":
        changed = jp.unpin()
    elif state == "enabled":
        changed = jp.enable()
    elif state == "disabled":
        changed = jp.disable()

    # Print status of the change
    module.exit_json(
        changed=changed,
        plugin=name,
        state=state,
        dependencies=jp.dependencies_states if hasattr(jp, "dependencies_states") else None,
    )


if __name__ == "__main__":
    main()



# Copyright (c) 2021, Ansible Team
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

ansible_version = '14.1.0'


#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: networks_floor_plans_auto_locate_jobs_batch
short_description: Resource module for networks _floor _plans _auto _locate _jobs
  _batch
description:
  - Manage operation create of the resource networks _floor _plans _auto _locate _jobs
    _batch.
  - Schedule auto locate jobs for one or more floor plans in a network.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module
author: Francisco Munoz (@fmunoz)
options:
  jobs:
    description: The list of auto locate jobs to be scheduled. Up to 100 jobs can
      be provided in a request.
    elements: dict
    suboptions:
      floorPlanId:
        description: The ID of the floor plan to run auto locate for.
        type: str
      refresh:
        description: The types of location data that should be refreshed for this
          job. The list must either contain both 'gnss' and 'ranging' or be empty,
          as we currently only support refreshing both 'gnss' and 'ranging', or neither.
        elements: str
        type: list
      scheduledAt:
        description: Timestamp in ISO8601 format which indicates when the auto locate
          job should be run. If omitted, the auto locate job will start immediately.
        type: str
    type: list
  networkId:
    description: NetworkId path parameter. Network ID.
    type: str
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for networks batchNetworkFloorPlansAutoLocateJobs
    description: Complete reference of the batchNetworkFloorPlansAutoLocateJobs API.
    link: https://developer.cisco.com/meraki/api-v1/#!batch-network-floor-plans-auto-locate-jobs
notes:
  - SDK Method used are
    networks.Networks.batch_network_floor_plans_auto_locate_jobs,
  - Paths used are
    post /networks/{networkId}/floorPlans/autoLocate/jobs/batch,
"""

EXAMPLES = r"""
- name: Create
  cisco.meraki.networks_floor_plans_auto_locate_jobs_batch:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    jobs:
      - floorPlanId: g_2176982374
        refresh:
          - gnss
          - ranging
        scheduledAt: '2018-02-11T00:00:00Z'
    networkId: string
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "jobs": [
        {
          "id": "string",
          "networkId": "string",
          "floorPlanId": "string",
          "status": "string",
          "scheduledAt": "string",
          "completed": {
            "percentage": 0
          },
          "ranging": {
            "status": "string",
            "completed": {
              "percentage": 0
            }
          },
          "gnss": {
            "status": "string",
            "completed": {
              "percentage": 0
            }
          },
          "errors": [
            {
              "source": "string",
              "type": "string"
            }
          ]
        }
      ]
    }
"""



#!/usr/bin/python
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = '''
---
module: azure_rm_ml_batch_endpoint_info
version_added: "3.18.0"
short_description: List ML Batch Endpoint resources
description:
    - List ML Batch Endpoint resources.
options:
    name:
        description:
            - Name of the Azure ML batch endpoint.
        type: str
        required: false
    resource_group:
        description:
            - Name of resource group.
        required: true
        type: str
    ml_workspace:
        description:
            - Name of the Azure ML workspace.
        required: true
        type: str

extends_documentation_fragment:
    - azure.azcollection.azure

author:
    - Bill Peck (@p3ck)
'''

EXAMPLES = '''
- name: Get Specific ML Batch Endpoint by name
  azure.azcollection.azure_rm_ml_batch_endpoint_info:
    name: endpoint-command-xxxxxxxxxx
    resource_group: myResourceGroup
    ml_workspace: myMLWorkspace

- name: List ML Batch Endpoints
  azure.azcollection.azure_rm_ml_batch_endpoint_info:
    resource_group: myResourceGroup
    ml_workspace: myMLWorkspace
'''

RETURN = '''
ml_batch_endpoints:
    description:
        - Batch Endpoints that match the query.
    returned: always
    type: dict
    sample: [
      {
          "auth_mode": "aad_token",
          "defaults": {
              "deployment_name": "deployment-xxxxxxxxxx-ansible"
          },
          "description": "A hello world endpoint for component deployments.",
          "id": "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/xxxxx-ml-batch_endpoint_invoke/providers/Microsoft.MachineLearningServices/workspaces/workspace-xxxxxxxxxx/batchEndpoints/endpoint-xxxxxxxxxx-ansible",
          "jobs": [
              {
                  "id": "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/xxxxx-ml-batch_endpoint_invoke/providers/Microsoft.MachineLearningServices/workspaces/workspace-xxxxxxxxxx/batchEndpoints/endpoint-xxxxxxxxxx-ansible/deployments/deployment-xxxxxxxxxx-ansible/jobs/pipelinejob-xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
                  "name": "pipelinejob-xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
                  "properties": {
                      "compute": {
                          "is_local": false
                      },
                      "interaction_endpoints": {
                          "Studio": {
                              "endpoint": "https://ml.azure.com/runs/pipelinejob-xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx?wsid=/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourcegroups/xxxxx-ml-batch_endpoint_invoke/workspaces/workspace-xxxxxxxxxx",
                              "job_endpoint_type": "Studio"
                          },
                          "Tracking": {
                              "endpoint": "azureml://eastus.api.azureml.ms/mlflow/v1.0/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/xxxxx-ml-batch_endpoint_invoke/providers/Microsoft.MachineLearningServices/workspaces/workspace-xxxxxxxxxx?",
                              "job_endpoint_type": "Tracking"
                          }
                      },
                      "name": "honest_grape_xxxxxxxx",
                      "output": {
                          "datastore_id": "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/xxxxx-ml-batch_endpoint_invoke/providers/Microsoft.MachineLearningServices/workspaces/workspace-xxxxxxxxxx/datastores/workspaceartifactstore",
                          "path": "ExperimentRun/dcid.xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx"
                      },
                      "output_data": {
                          "score": {
                              "job_output_type": "UriFile",
                              "mode": "ReadWriteMount",
                              "uri": "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/xxxxx-ml-batch_endpoint_invoke/providers/Microsoft.MachineLearningServices/workspaces/workspace-xxxxxxxxxx/datastores/workspaceblobstore/paths/azureml/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/score/predictions.csv"
                          }
                      },
                      "properties": {
                          "azureml.DatasetAccessMode": "Asset",
                          "azureml.DevPlatv2": "true",
                          "azureml.SourceComponentId": "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/xxxxx-ml-batch_endpoint_invoke/providers/Microsoft.MachineLearningServices/workspaces/workspace-xxxxxxxxxx/components/hello_batch/versions/1",
                          "azureml.continue_on_failed_optional_input": "True",
                          "azureml.continue_on_step_failure": "True",
                          "azureml.defaultComputeName": "batch-cluster",
                          "azureml.defaultDataStoreName": "workspaceblobstore",
                          "azureml.deploymentname": "deployment-xxxxxxxxxx-ansible",
                          "azureml.endpointname": "endpoint-xxxxxxxxxx-ansible",
                          "azureml.enforceRerun": "False",
                          "azureml.parameters": "{}",
                          "azureml.pipelineComponent": "pipelinerun",
                          "azureml.pipelines.stages": "{\\"Initialization\\":null,\\"Execution\\":{\\"StartTime\\":\\"2026-05-04T13:51:04.3564364+00:00\\",\\"EndTime\\":\\"2026-05-04T13:55:32.6682006+00:00\\",\\"Status\\":\\"Finished\\"}}",
                          "azureml.runLineageType": "batchDeployment",
                          "azureml.runsource": "azureml.PipelineRun",
                          "azureml.sourcepipelinerunid": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
                          "runSource": "BatchDeployment",
                          "runType": "HTTP"
                      },
                      "provisioning_state": "Succeeded",
                      "status": "Completed",
                      "tags": {
                          "azureml.batchrun": "true",
                          "azureml.deploymentname": "deployment-xxxxxxxxxx-ansible",
                          "azureml.jobtype": "azureml.pipelinejob"
                      }
                  },
                  "system_data": {
                      "created_at": "2026-05-04T13:51:03.171352Z",
                      "created_by": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
                      "last_modified_at": "2026-05-04T13:55:32.837426Z",
                      "last_modified_by": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx"
                  },
                  "type": "Microsoft.MachineLearningServices/workspaces/batchEndpoints/deployments/jobs"
              }
          ],
          "location": "eastus",
          "name": "endpoint-xxxxxxxxxx-ansible",
          "properties": {
              "BatchEndpointCreationApiVersion": "2023-10-01",
              "azureml.onlineendpointid": "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/xxxxx-ml-batch_endpoint_invoke/providers/Microsoft.MachineLearningServices/workspaces/workspace-xxxxxxxxxx/batchEndpoints/endpoint-xxxxxxxxxx-ansible"
          },
          "provisioning_state": "Succeeded",
          "scoring_uri": "https://endpoint-xxxxxxxxxx-ansible.eastus.inference.ml.azure.com/jobs",
          "tags": {}
      }
    ]
'''  # NOQA


try:
    from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common_ml import MLClientCommon
    from azure.core.exceptions import ResourceNotFoundError
except ImportError:
    # This is handled in azure_rm_common
    pass


class AzureRMMLBatchEndpointInfo(MLClientCommon):
    def __init__(self):

        self.module_arg_spec = dict(
            name=dict(
                type='str',
                required=False,
            ),
            resource_group=dict(
                type='str',
                required=True,
            ),
            ml_workspace=dict(
                type='str',
                required=True,
            ),
        )

        self._client = None
        self.ml_registry = None

        self.results = dict(
            ml_batch_endpoints=[]
        )

        super(AzureRMMLBatchEndpointInfo, self).__init__(self.module_arg_spec,
                                                         supports_tags=False,
                                                         supports_check_mode=True,
                                                         facts_module=True,
                                                         )

    def exec_module(self, **kwargs):

        for key in list(self.module_arg_spec.keys()):
            setattr(self, key, kwargs[key])

        if self.name:
            ml_batch_endpoints = [self.get(self.name)]
        else:
            results = self.client.batch_endpoints.list()
            ml_batch_endpoints = [self.get(x.name) for x in results]

        self.results['ml_batch_endpoints'] = ml_batch_endpoints
        return self.results

    def get(self, name):
        try:
            result = self.client.batch_endpoints.get(name=name)
            ml_batch_endpoint = self.entity_to_dict(result)
            jobs = [self.entity_to_dict(x) for x in self.client.batch_endpoints.list_jobs(endpoint_name=name)]
            ml_batch_endpoint['jobs'] = jobs
        except ResourceNotFoundError:
            ml_batch_endpoint = None

        return ml_batch_endpoint


def main():
    AzureRMMLBatchEndpointInfo()


if __name__ == '__main__':
    main()
